import gzip
import json
import logging
import os
import osometweet
import re
import sys
import time
import numpy as np
import scipy.sparse as sp

from datetime import datetime

from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.preprocessing import normalize

logger = logging.getLogger('bloc.bloc')

def procLogHandler(handler, loggerDets):
    
    if( handler is None ):
        return
        
    if( 'level' in loggerDets ):
        handler.setLevel( loggerDets['level'] )    
        
        if( loggerDets['level'] == logging.ERROR ):
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s :\n%(message)s')
            handler.setFormatter(formatter)

    if( 'format' in loggerDets ):
        
        loggerDets['format'] = loggerDets['format'].strip()
        if( loggerDets['format'] != '' ):
            formatter = logging.Formatter( loggerDets['format'] )
            handler.setFormatter(formatter)

    logger.addHandler(handler)

def setLoggerDets(logger, loggerDets):

    if( len(loggerDets) == 0 ):
        return

    consoleHandler = logging.StreamHandler()

    if( 'level' in loggerDets ):
        logger.setLevel( loggerDets['level'] )
    else:
        logger.setLevel( logging.INFO )

    if( 'file' in loggerDets ):
        loggerDets['file'] = loggerDets['file'].strip()
        
        if( loggerDets['file'] != '' ):
            fileHandler = logging.FileHandler( loggerDets['file'] )
            procLogHandler(fileHandler, loggerDets)

    procLogHandler(consoleHandler, loggerDets)

def setLogDefaults(params):
    
    params['log_dets'] = {}

    if( params['log_level'] == '' ):
        params['log_dets']['level'] = logging.INFO
    else:
        
        logLevels = {
            'CRITICAL': 50,
            'ERROR': 40,
            'WARNING': 30,
            'INFO': 20,
            'DEBUG': 10,
            'NOTSET': 0
        }

        params['log_level'] = params['log_level'].strip().upper()

        if( params['log_level'] in logLevels ):
            params['log_dets']['level'] = logLevels[ params['log_level'] ]
        else:
            params['log_dets']['level'] = logging.INFO
    
    params['log_format'] = params['log_format'].strip()
    params['log_file'] = params['log_file'].strip()

    if( params['log_format'] != '' ):
        params['log_dets']['format'] = params['log_format']

    if( params['log_file'] != '' ):
        params['log_dets']['file'] = params['log_file']

def genericErrorInfo(slug=''):
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    
    errMsg = fname + ', ' + str(exc_tb.tb_lineno)  + ', ' + str(sys.exc_info())
    logger.error(errMsg + slug)

    return errMsg

#http://stackoverflow.com/questions/4770297/python-convert-utc-datetime-string-to-local-datetime
def datetimeFromUtcToLocal(utc_datetime):
    now_timestamp = time.time()
    offset = datetime.fromtimestamp(now_timestamp) - datetime.utcfromtimestamp(now_timestamp)
    return utc_datetime + offset

#tweet_datetime: 2020-08-10T23:59:59
def gen_post_snowflake_twitter_id(tweet_datetime):
    
    pre_snowflake_last_tweet_id = 29700859247 #(2010-11-04T21:02:48.000Z)

    offset = 1288834974657
    
    if( isinstance(tweet_datetime, str) ):
        tweet_datetime = datetime.strptime(tweet_datetime, '%Y-%m-%d %H:%M:%S')
    
    tweet_datetime = (tweet_datetime.timestamp() * 1000) - offset#convert seconds to nanoseconds by * 1000
    tweet_datetime = int(tweet_datetime)

    tweet_id = "{0:b}".format(tweet_datetime) + '1'*22
    tweet_id = int(tweet_id, 2)

    return tweet_id

def find_tweet_timestamp_post_snowflake(tid):
    #credit: https://github.com/oduwsdl/tweetedat/blob/master/script/TimestampEstimator.py#L175 and https://ws-dl.blogspot.com/2019/08/2019-08-03-tweetedat-finding-tweet.html
    offset = 1288834974657
    tstamp = (tid >> 22) + offset
    utcdttime = datetime.utcfromtimestamp(tstamp/1000)
    return utcdttime

#extract realDonaldTrump from 'https://twitter.com/realDonaldTrump/status/863007411132649473'
def get_screen_name_frm_status_uri(tweet_uri):

    tweet_uri = tweet_uri.strip()
    tweet_uri = tweet_uri.replace('http://twitter.com/', 'https://twitter.com/')

    if( tweet_uri.startswith('https://twitter.com/') is False or tweet_uri.find('/status/') == -1 ):
        return ''

    return tweet_uri.replace('https://twitter.com/', '').split('/')[0]
#file - start

def readTextFromFile(infilename):

    text = ''

    try:
        with open(infilename, 'r') as infile:
            text = infile.read()
    except:
        print('\treadTextFromFile()error filename:', infilename)
        genericErrorInfo()
    

    return text

def dumpJsonToFile(outfilename, dictToWrite, indentFlag=True, extraParams=None):

    if( extraParams is None ):
        extraParams = {}

    extraParams.setdefault('verbose', True)

    try:
        outfile = open(outfilename, 'w')
        
        if( indentFlag ):
            json.dump(dictToWrite, outfile, ensure_ascii=False, indent=4)#by default, ensure_ascii=True, and this will cause  all non-ASCII characters in the output are escaped with \uXXXX sequences, and the result is a str instance consisting of ASCII characters only. Since in python 3 all strings are unicode by default, forcing ascii is unecessary
        else:
            json.dump(dictToWrite, outfile, ensure_ascii=False)

        outfile.close()

        if( extraParams['verbose'] ):
            logger.info('\tdumpJsonToFile(), wrote: ' + outfilename)
    except:
        genericErrorInfo('\n\terror: outfilename: ' + outfilename)
        return False

    return True

def getDictFromJsonGZ(path):

    json = getTextFromGZ(path)
    if( len(json) == 0 ):
        return {}
    return getDictFromJson(json)

def getTextFromGZ(path):
    
    try:
        with gzip.open(path, 'rb') as f:
            return f.read().decode('utf-8')
    except:
        genericErrorInfo()

    return ''

def getDictFromJson(jsonStr):

    try:
        return json.loads(jsonStr)
    except:
        genericErrorInfo('\tjsonStr prefix: ' + jsonStr[:100])

    return {}

def getDictFromFile(filename):

    try:

        if( os.path.exists(filename) == False ):
            return {}

        return getDictFromJson( readTextFromFile(filename) )
    except:
        print('\tgetDictFromFile(): error filename', filename)
        genericErrorInfo()

    return {}

def gzipTextFile(path, txt):
    
    try:
        with gzip.open(path, 'wb') as f:
            f.write(txt.encode())
    except:
        genericErrorInfo()

#file - end

#/stringmatrix/vector manipulation - start

def get_color_txt(txt, ansi_code='91m'):
    return txt if txt.strip() == '' else '\033[' + ansi_code + '{}\033[00m'.format(txt)

def color_bloc_action_str(bloc_action_str, split_pattern='([^□⚀⚁⚂⚃⚄⚅.|()*]+)', ansi_code='91m'):

    if( ansi_code.strip() == '' ):
        return bloc_action_str

    colored_bloc_action_str = ''
    bloc_action_tokens = re.split(split_pattern, bloc_action_str)
    
    for tok in bloc_action_tokens:

        if( tok.strip() not in split_pattern ):
            tok = get_color_txt(tok, ansi_code=ansi_code)
        
        colored_bloc_action_str += tok

    return colored_bloc_action_str

def get_char_ngram(string, n):

    if( n < 1 or n > len(string) ):
        return []

    return [ string[i:i+n] for i in range(len(string)-n+1) ]

def get_char_ngram_vocab(string, n):
    
    ngrams = get_char_ngram(string, n)
    vocab = {}
    
    for gram in ngrams:

        vocab.setdefault(gram, 0)
        vocab[gram] += 1

    vocab = sorted(vocab.items(), key=lambda x: x[1], reverse=True)
    return vocab

def get_doc_lst_pos_maps(doc_lst, rm_text=True):

    new_doc_list = []
    pos_id_mapping = {}

    try:

        for i in range( len(doc_lst) ):

            d = doc_lst[i]
            d.setdefault('id', i)
            pos_id_mapping[i] = {'id': d['id']}

            #transfer other properties - start
            for ky, val in d.items():

                if( rm_text is True ):
                    if( ky == 'text' ):
                        continue

                pos_id_mapping[i][ky] = val
            #transfer other properties - end

            new_doc_list.append( d['text'] )
    
    except:
        genericErrorInfo()
        return [], {}

    return new_doc_list, pos_id_mapping

def map_tf_mat_to_doc_ids(payload, pos_id_mapping):

    #see get_tf_matrix().payload for payload's structure
    if( 'tf_matrix' not in payload and 'top_ngrams' not in payload ):
        return {}

    if( 'per_doc' not in payload['top_ngrams'] ):
        return {}

    tf_matrix = []
    top_ngrams_per_doc = []
    for pos, doc_dct in pos_id_mapping.items():
        if( pos < len(payload['tf_matrix']) ):

            tf_matrix.append({
                'id': doc_dct['id'], 
                'tf_vector': payload['tf_matrix'][pos]
            })

            if( len(payload['top_ngrams']['per_doc']) != 0 ):
                top_ngrams_per_doc.append({
                    'id': doc_dct['id'],
                    'ngrams': payload['top_ngrams']['per_doc'][pos]
                })

            #transfer other properties - start
            for ky, val in doc_dct.items():
                tf_matrix[-1][ky] = val

                if( len(top_ngrams_per_doc) != 0 ):
                    top_ngrams_per_doc[-1][ky] = val
            #transfer other properties - end


    #special case for (optional, see: get_tf_matrix().tf_matrix_norm) tf_matrix_normalized - start
    for opt in ['tf_matrix_normalized', 'tf_idf_matrix']:
        if( opt in payload ):
            
            opt_vect = []
            for pos, doc_dct in pos_id_mapping.items():
                if( pos < len(payload[opt]) ):

                    opt_vect.append({
                        'id': doc_dct['id'], 
                        'tf_vector': payload[opt][pos]
                    })

                    #transfer other properties - start
                    for ky, val in doc_dct.items():
                        opt_vect[-1][ky] = val
                    #transfer other properties - end

            payload[opt] = opt_vect
    #special case for (optional, see: get_tf_matrix().tf_matrix_norm) tf_matrix_normalized - end

    payload['tf_matrix'] = tf_matrix
    payload['top_ngrams']['per_doc'] = top_ngrams_per_doc

    return payload

def gen_folded_vocab( tf_matrix, bloc_variant ):

    if( 'vocab' not in tf_matrix or 'type' not in bloc_variant ):
        return {}

    expected_types = ['folded_words']
    if( bloc_variant['type'] not in expected_types ):
        return {}
    
    new_vocab = {}
    logger.debug('\ngen_folded_vocab():')
    logger.debug('\tfold_start_count:', bloc_variant['fold_start_count'])

    for i in range( len(tf_matrix['vocab']) ):
        
        v = tf_matrix['vocab'][i]
        new_v = fold_word(v, bloc_variant['fold_start_count'], count_applies_to_all_char=bloc_variant['count_applies_to_all_char'], fold_exclude_chars=bloc_variant.get('fold_exclude_chars', '') )

        new_vocab.setdefault(new_v, [])
        new_vocab[new_v].append(i)

    return new_vocab

def get_default_symbols():
    bloc_symbols_file = '{}/symbols.json'.format(os.path.dirname(os.path.abspath(__file__)))
    return getDictFromFile(bloc_symbols_file)

def gen_bloc_variant_tf_mat( tf_matrix, bloc_variant ):

    if( 'tf_matrix' not in tf_matrix or 'vocab' not in tf_matrix or 'type' not in bloc_variant ):
        return {}

    if( len(tf_matrix['tf_matrix']) == 0 ):
        return {}

    expected_types = ['folded_words']
    if( bloc_variant['type'] not in expected_types ):
        return {}
    
    new_vocab = {}
    bloc_variant.setdefault('fold_start_count', 4)
    bloc_variant.setdefault('count_applies_to_all_char', False)
    bloc_variant.setdefault('fold_exclude_chars', '')

    logger.debug( '\ngen_bloc_variant_tf_mat():' )
    logger.debug( '\ttype: {}'.format(bloc_variant['type']) )
    logger.debug( '\tfold_start_count: {}'.format(bloc_variant['fold_start_count']) )
    logger.debug( '\tcount_applies_to_all_char: {}'.format(bloc_variant['count_applies_to_all_char']) )
    logger.debug( '\told vocab: {} entries'.format(len(tf_matrix['vocab'])) )

    tf_mat = tf_matrix['tf_matrix'][0]['tf_vector']
    for i in range( 1, len(tf_matrix['tf_matrix']) ):
        tf_mat = sp.vstack( (tf_mat, tf_matrix['tf_matrix'][i]['tf_vector']), format='csr' )

    new_vocab = gen_folded_vocab( tf_matrix, bloc_variant )

    logger.debug('')
    logger.debug( f'new_vocab: {len(new_vocab)} entries' )
    
    feature_indx = 0
    new_tf_mat = None
    new_vocab_lst = []
    #this function merges columns of features across all users
    for v, col_indices in new_vocab.items():
        
        logger.debug( '\tfeature_indx: {} {} col_indices: {}'.format(feature_indx, v, col_indices) )
        feature_indx += 1
        new_vocab_lst.append(v)
        
        if( len(col_indices) == 1 ):
            
            indx = col_indices[0]
            if( new_tf_mat is None ):
                new_tf_mat = tf_mat[:,indx]
            else:
                new_tf_mat = sp.hstack( (new_tf_mat, tf_mat[:,indx]), format='csr' )
        else:
            
            #sum all columns (from col_indices) in tf_mat to merge features
            indx = col_indices[0]
            col_sum = tf_mat[:,indx]
            col_indices.pop(0)

            for indx in col_indices:
                col_sum = np.sum( [col_sum, tf_mat[:,indx]], axis=0 )

            if( new_tf_mat is None ):
                new_tf_mat = col_sum
            else:
                new_tf_mat = sp.hstack( (new_tf_mat, col_sum), format='csr' )

    logger.debug('')
    logger.debug( '\tnew_tf_mat shape: {}'.format(new_tf_mat.shape) )
    logger.debug( '\tnew_vocab: {}'.format(new_vocab_lst) )
    logger.debug('')

    return {
        'tf_matrix': new_tf_mat,
        'vocab': new_vocab_lst
    }

def update_bloc_model( old_bloc_model, new_tf_mat, tf_matrix_norm, tf_idf_norm, **kwargs ):
    
    if( len(new_tf_mat) == 0 ):
        return old_bloc_model
    
    #mimic signature of function call in gen_ngram_tf_matrix() and process payload similar to gen_ngram_tf_matrix()
    pos_id_mapping = kwargs.pop('pos_id_mapping', None)
    new_tf_mat = get_tf_matrix( 
        doc_lst=['dummy'], 
        n=0, 
        tf_mat=new_tf_mat['tf_matrix'],
        vocab=new_tf_mat['vocab'],
        tf_matrix_norm=tf_matrix_norm,
        tf_idf_norm=tf_idf_norm,
        pos_id_mapping=None,
        **kwargs
    )

    if( pos_id_mapping is not None ):
        kwargs['pos_id_mapping'] = pos_id_mapping

    if( 'token_pattern' in old_bloc_model ):
        new_tf_mat['token_pattern'] = old_bloc_model['token_pattern']


    for m in ['tf_matrix', 'tf_matrix_normalized', 'tf_idf_matrix']:
        
        if( m not in new_tf_mat ):
            continue

        for i in range( len(new_tf_mat[m]) ):
            
            old_bloc_model[m][i]['tf_vector'] = new_tf_mat[m][i]
            new_tf_mat[m][i] = old_bloc_model[m][i]


    for m in ['per_doc', 'all_docs']:

        if( 'top_ngrams' not in new_tf_mat or 'top_ngrams' not in old_bloc_model ):
            continue

        if( m not in new_tf_mat['top_ngrams'] or m not in old_bloc_model['top_ngrams'] ):
            continue

        if( len(new_tf_mat['top_ngrams'][m]) != len(old_bloc_model['top_ngrams'][m]) ):
            continue

        for i in range( len(new_tf_mat['top_ngrams'][m]) ):
            
            old_bloc_model['top_ngrams'][m][i]['ngrams'] = new_tf_mat['top_ngrams'][m][i]
            new_tf_mat['top_ngrams'][m][i] = old_bloc_model['top_ngrams'][m][i]

    return new_tf_mat

def get_bloc_lite_twt_frm_full_twt(tweet):
    
    def tranfer_dets_for_stream_statuses(twt):
        
        #transfer details for tweets gotten from streams
        #See: http://web.archive.org/web/20210812135100/https://docs.tweepy.org/en/stable/extended_tweets.html
        if( 'extended_tweet' not in twt ):
            return twt

        for ky in ['full_text', 'display_text_range', 'entities', 'extended_entities']:
            if( ky in twt['extended_tweet'] ):
                twt[ky] = twt['extended_tweet'][ky]

        return twt

    if( 'id' not in tweet ):
        return {}

    tweet = tranfer_dets_for_stream_statuses(tweet)
    payload = {
        'id': tweet['id'],
        'source': tweet['source'],
        'created_at': tweet['created_at'],
        'user': tweet['user']
    }
    
    for itm in ['full_text', 'text']:
        if( itm in tweet ):
            payload['full_text'] = tweet[itm]
            break
    

    if( tweet['in_reply_to_status_id'] is None ):
        payload['in_reply_to_status_id'] = None
        payload['in_reply_to_user_id'] = None
    else:
        payload['in_reply_to_status_id'] = tweet['in_reply_to_status_id']
        payload['in_reply_to_user_id'] = tweet['in_reply_to_user_id']

    
    payload['in_reply_to_screen_name'] = tweet['in_reply_to_screen_name']
    for itm in ['display_text_range', 'entities', 'extended_entities']:
        if( itm in tweet ):
            payload[itm] = tweet[itm]


    if( 'retweeted_status' in tweet ):
        payload['retweeted_status'] = {
            'user': {'screen_name': tweet['retweeted_status']['user']['screen_name'], 'id': tweet['retweeted_status']['user']['id']},
            'created_at': tweet['retweeted_status']['created_at'],
            'id': tweet['retweeted_status']['id']
        }
    
    return payload

def get_bloc_doc_lst(twts_bloc, dimensions, src='', src_class=''):

    doc_lst = []

    for d in twts_bloc:

        if( 'bloc' not in d ):
            continue
        
        doc = [ d['bloc'][dim] for dim in dimensions if dim in d['bloc'] ]
        if( len(doc) == 0 ):
            continue

        doc = ''.join(doc)
        doc = doc.strip()
    
        if( doc == '' ):
            continue

        doc_lst.append({
            'text': doc,
            'user_id': d['user_id'],
            'screen_name': d['screen_name'],
            'src': src,
            'class': src_class
        })

    return doc_lst

def get_bloc_variant_tf_matrix(doc_lst, ngram, tf_mat=None, vocab=None, token_pattern='[^□⚀⚁⚂⚃⚄⚅. |()*]+|[□⚀⚁⚂⚃⚄⚅.]', **kwargs):

    #bigram: r'[^ |()*]', word: '[^□⚀⚁⚂⚃⚄⚅. |()*]+|[□⚀⚁⚂⚃⚄⚅.]'
    #reconcile implementation with bloc_analyzer.py::analyze_bloc_for_users()
    bloc_variant = kwargs.get('bloc_variant', None)
    if( doc_lst == [] ):
       return {}

    if( bloc_variant is None ):

        tf_matrix = get_tf_matrix( 
            doc_lst, 
            ngram,
            tf_mat=tf_mat,
            vocab=vocab,
            token_pattern=token_pattern,
            lowercase=False, 
            **kwargs
        )

    else:
        tf_matrix = get_tf_matrix( 
            doc_lst, 
            ngram,
            tf_mat=tf_mat,
            vocab=vocab,
            token_pattern=token_pattern,
            lowercase=False,
            keep_tf_matrix=True,
            tf_matrix_norm='',
            tf_idf_norm='',
            pos_id_mapping=kwargs.get('pos_id_mapping', None)
        )

        bloc_variant_tf_matrix = gen_bloc_variant_tf_mat( tf_matrix, bloc_variant )
        
        #since tf_matrix only has tf_matrix, transfer properties for each document to tf_matrix so they'd eventually be transfered to bloc_variant_tf_matrix - start
        for m in ['tf_matrix_normalized', 'tf_idf_matrix']:

            if( m not in tf_matrix ):
                continue

            for i in range( len(tf_matrix['tf_matrix']) ):
                doc_dct = [ (v[0], v[1]) for v in tf_matrix['tf_matrix'][i].items() if v[0] != 'tf_vector' ]
                tf_matrix[m].append( dict(doc_dct) )

        tf_matrix = update_bloc_model(tf_matrix, bloc_variant_tf_matrix, tf_matrix_norm=kwargs.pop('tf_matrix_norm', ''), tf_idf_norm=kwargs.pop('tf_idf_norm', 'l2'), **kwargs )
        if( kwargs.get('keep_tf_matrix', False) is False ):
            tf_matrix['tf_matrix'] = []
    
    return tf_matrix

def conv_tf_matrix_to_json_compliant(tf_mat):
    
    if( 'vocab' in tf_mat ):
        tf_mat['vocab'] = list(tf_mat['vocab'])

    for opt in ['tf_matrix', 'tf_matrix_normalized', 'tf_idf_matrix']:
        if( opt not in tf_mat ):
            continue
        
        for i in range( len(tf_mat[opt]) ):
            tf_mat[opt][i]['tf_vector'] = [ float(a) for a in tf_mat[opt][i]['tf_vector'].toarray()[0] ]

    return tf_mat

def get_tf_matrix(doc_lst, n, tf_mat=None, vocab=None, token_pattern=r'(?u)\b[a-zA-Z\'\’-]+[a-zA-Z]+\b|\d+[.,]?\d*', **kwargs):

    kwargs.setdefault('rm_doc_text', True)
    pos_id_mapping = kwargs.get('pos_id_mapping', None)

    if( isinstance(doc_lst, list) ):
        if( len(doc_lst) == 0 ):
            return {}

        if( isinstance(doc_lst[0], dict) ):
            doc_lst, pos_id_mapping = get_doc_lst_pos_maps( doc_lst, rm_text=kwargs['rm_doc_text'] )
            if( len(doc_lst) == 0 ):
                return {}
    
    
    kwargs.setdefault('top_ngrams_add_all_docs', False)
    kwargs.setdefault('keep_tf_matrix', True)
    kwargs.setdefault('lowercase', True)
    kwargs.setdefault('min_df', 1)
    kwargs.setdefault('set_top_ngrams', False)
    kwargs.setdefault('tf_matrix_norm', '')#can use l1
    kwargs.setdefault('tf_idf_norm', 'l2')
    kwargs.setdefault('count_vectorizer_kwargs', {})

    
    #if payload changes, update map_tf_mat_to_doc_ids()
    #also update conv_tf_matrix_to_json_compliant()
    payload = {
        'tf_matrix': [],
        'tf_matrix_normalized': [],
        'tf_idf_matrix': [],
        'vocab': [],
        'top_ngrams': {
            'per_doc': [],
            'all_docs': []
        },
        'token_pattern': token_pattern
    }
    
    try:
        count_vectorizer = CountVectorizer(
            stop_words=kwargs.get('stop_words', None), 
            tokenizer=kwargs.get('tokenizer', None),
            token_pattern=token_pattern,
            ngram_range=(n, n),
            lowercase=kwargs['lowercase'],
            min_df=kwargs['min_df'],
            **kwargs['count_vectorizer_kwargs']
        )

        if( tf_mat is not None and vocab is not None ):
            payload['tf_matrix'] = tf_mat
            payload['vocab'] = vocab
        else:
            payload['tf_matrix'] = count_vectorizer.fit_transform(doc_lst)
            payload['vocab'] = count_vectorizer.get_feature_names_out()

        if( kwargs['tf_matrix_norm'] != '' ):
            payload['tf_matrix_normalized'] = normalize(payload['tf_matrix'], norm=kwargs['tf_matrix_norm'], axis=1)

        if( kwargs['tf_idf_norm'] != '' ):
            tfidf = TfidfTransformer( norm=kwargs['tf_idf_norm'] )
            tfidf.fit(payload['tf_matrix'])
            payload['tf_idf_matrix'] = tfidf.transform(payload['tf_matrix'])

        if( kwargs['keep_tf_matrix'] is False ):
            payload['tf_matrix'] = []

        
        for opt in ['tf_matrix', 'tf_matrix_normalized', 'tf_idf_matrix']:
            if( opt not in payload ):
                continue
            payload[opt] = list( payload[opt] )
    except:
        genericErrorInfo()
        return {}
    
    if( kwargs['set_top_ngrams'] is True ):
        calc_top_ngrams( payload, top_ngrams_add_all_docs=kwargs['top_ngrams_add_all_docs'] )

    if( pos_id_mapping is not None ):
        payload = map_tf_mat_to_doc_ids(payload, pos_id_mapping)

    return payload

def calc_top_ngrams(payload, top_ngrams_add_all_docs=False):

    all_docs_tf = {}
    all_docs_total_tf = 0
    corpus_size = len(payload['tf_matrix'])
    for i in range( corpus_size ):
        
        row = payload['tf_matrix'][i].toarray()[0]
        total_tf = sum(row)
        if( total_tf == 0 ):
            total_tf = -1

        single_doc_tf = [ {'term': v, 'term_freq': int(row[j]), 'term_rate': row[j]/total_tf} for (j, v) in enumerate(payload['vocab']) if row[j] != 0 ]
        single_doc_tf = sorted( single_doc_tf, key=lambda i: i['term_freq'], reverse=True )
        
        payload['top_ngrams']['per_doc'].append( single_doc_tf )


        if( top_ngrams_add_all_docs is True ):
            for tf_dct in single_doc_tf:
                all_docs_tf.setdefault( tf_dct['term'], {'tf': 0, 'df': 0} )
                all_docs_tf[ tf_dct['term'] ]['tf'] += int(tf_dct['term_freq'])
                all_docs_tf[ tf_dct['term'] ]['df'] += 1
                all_docs_total_tf += int(tf_dct['term_freq'])

    
    if( top_ngrams_add_all_docs is True ):

        if( all_docs_total_tf == 0 ):
            all_docs_total_tf = -1

        payload['top_ngrams']['all_docs'] = sorted( all_docs_tf.items(), key=lambda x: x[1]['df'], reverse=True )
        payload['top_ngrams']['all_docs'] = [ {'term': t[0], 'term_freq': t[1]['tf'], 'term_rate': t[1]['tf']/all_docs_total_tf, 'doc_freq': t[1]['df'], 'doc_rate': t[1]['df']/corpus_size} for t in payload['top_ngrams']['all_docs'] ]

def segment_paren_word(word, sort_paren=True):
    word = word.strip()
    if( word == '' ):
        return []

    paren_words = ['']
    for w in word:

        if( w == '(' ):
            paren_words.append('')
            continue
        elif( w == ')' ):
            paren_words[-1] = '(' + ''.join( sorted(paren_words[-1]) ) + ')' if sort_paren is True else '(' + paren_words[-1] + ')'
            paren_words.append('')
            continue

        paren_words[-1] += w

    paren_words = [w for w in paren_words if w != '']
    return paren_words

def gen_bloc_folded_letters(word, fold_start_count=100, sort_paren=True):

    if( word == '' ):
        return []
    
    '''
        gen_bloc_folded_letters() examples

        E:   E-1         [(E, 1)]
        EE:  E-2         [(E, 2)]
        EEAE:E-2,A-1,E-1 [(E, 2), (A, 1), (E, 1)]
        AB:  A-1,B-1     [(A, 1), (B, 1)]
    '''
    if( word.find('(') == -1 ):
        #e.g., word: AAAAABBBBBBABABABAB
        word = [ w for w in word ]
    else:
        '''
        e.g., word: (AAAAABBBBBB)(AB)(AB)(AB)(AB) -> (AAA+BBB+)(AB)(AB)(AB)+, fold_start_count = 4
        without recursive call to fold_word(), word: (AAAAABBBBBB)(AB)(AB)(AB)(AB) -> word: (AAAAABBBBBB)(AB)(AB)(AB)+
        '''
        word = segment_paren_word(word, sort_paren)

        new_word = []
        for w in word:
            if( w.startswith('(') and w.endswith(')') ):
                new_word.append( '(' + fold_word(w[1:-1], fold_start_count) + ')' )
            else:
                new_word.append( fold_word(w, fold_start_count) )
        word = new_word
        
    c_count = {}
    prev_c = word[0]
    word = word + ['*']
    folded_word = []
    
    for i in range( len(word) ):
        
        c = word[i]

        if( prev_c != c ):
            freq = c_count[prev_c]['freq']
            folded_word.append( (prev_c, freq) )
            c_count = {}

        c_count.setdefault(c, {'freq': 0, 'pos': i})
        c_count[c]['freq'] += 1
        prev_c = c

    return folded_word

def fold_word(word, fold_start_count, sort_paren=True, count_applies_to_all_char=False, fold_exclude_chars=''):

    if( fold_start_count < 1 ):
        return word

    if( isinstance(word, str) ):
        word = gen_bloc_folded_letters(word, fold_start_count=fold_start_count, sort_paren=sort_paren)

    new_word = ''
    exceed_cf_count = 0
    create_vocab_new_entry = True

    if( count_applies_to_all_char is True ):

        for c, cf in word:
            if( cf >= fold_start_count and c not in fold_exclude_chars ):
                
                plus_count = fold_start_count - 1 if fold_start_count > 1 else 1
                
                new_word += f'{c*plus_count}+'
                exceed_cf_count += 1
            else:
                new_word += f'{c*cf}'

        if( exceed_cf_count == len(word) ):
            create_vocab_new_entry = False
        else:
            new_word  = ''.join([ c[0] for c in word ])
    else:

        for c, cf in word:
            if( cf >= fold_start_count and c not in fold_exclude_chars ):
                
                create_vocab_new_entry = False
                plus_count = fold_start_count - 1 if fold_start_count > 1 else 1
                
                new_word += f'{c*plus_count}+'
            else:
                new_word += f'{c*cf}'

    return new_word


def bloc_sf_content_map(bloc_str, typ):

    sf = ''
    bloc_str = bloc_str.strip()
    if( bloc_str == '' ):
        return ''
    
    if( typ == 'b3_content' ):
        if( bloc_str == 't' ):
            sf = 'N' #N: tweet contains no entities (plaintext)
        elif( len(set(bloc_str)) == 1 ):
            sf = 'E' #E: tweet contains entities of one type
        else:
            sf = 'X' #X: tweet contains entities of mixed types

        return sf
    

    bloc_sf_b6_content_map = {
        't': 'N', #N: tweet contains no entities (plaintext)
        'U': 'U', #U: tweet contains 1+ URLs
        'H': 'H', #H: tweet contains 1+ Hashtags
        'm': 'M', #M: tweet contains 1+ Mentions
        'M': 'M', #M: tweet contains 1+ Mentions
        'E': 'D'  #D: tweet contains 1+ Media
    }
    
    ori = bloc_str
    bloc_str = ''.join( set(bloc_str) )
    if( len(bloc_str) > 1 ):
        sf = 'X'
    else:
        sf = bloc_sf_b6_content_map.get(bloc_str, '')
    
    return sf

def get_social_fingerprint_frm_bloc( bloc_str, dimension='action' ):
    #social fingerprint based on: https://ieeexplore.ieee.org/abstract/document/7876716    
    

    if( dimension == 'action' ):
        sf = re.sub(r'[^Tpr]', '', bloc_str)
        sf = re.sub(r'[T]', 'A', sf) #tweet 
        sf = re.sub(r'[p]', 'C', sf) #reply
        sf = re.sub(r'[r]', 'T', sf) #retweet
        return [{ 'text': sf, 'type': 'b3_type' }]

    elif( dimension == 'content_syntactic' ):
        sf = re.sub(r'[ (|*]', '', bloc_str)
        sf = sf.split(')')
        sf = [ s for s in sf if s != '' ]
        
        b3_c = [ bloc_sf_content_map(s, 'b3_content') for s in sf ]
        b6_c = [ bloc_sf_content_map(s, 'b6_content') for s in sf ]
        
        return [
            {'text': ''.join(b3_c), 'type': 'b3_content'},
            {'text': ''.join(b6_c), 'type': 'b6_content'}
        ]
    
    return []

#/stringmatrix/vector manipulation - end

#twitter v2 - start
def twitter_v2_user_lookup_ids(osome_twt_obj, screen_names):

    logger.info( '\ntwitter_v2_user_lookup_ids():' )

    max_users_for_req = len(screen_names)
    all_users = []
    
    for i in range(0, max_users_for_req, 100):
        
        logger.info(f'\tscreen_name lookup {i+1} of {max_users_for_req//100}, total users: {max_users_for_req}')
        
        if( max_users_for_req > 100 ):
            logger.info('\t\tsleeping for 3 seconds, check if 1 seconds would do')
            time.sleep(3)
 
        try:
            response = osome_twt_obj.user_lookup_usernames( screen_names[i:i+100], fields=osometweet.UserFields(everything=True) )
        except:
            genericErrorInfo()
            continue
        
        all_users += response.get('data', [])

    return all_users

#twitter v2 - end



import logging
from itertools import combinations
from sklearn.metrics.pairwise import cosine_similarity

from bloc.generator import add_bloc_sequences
from bloc.generator import get_word_type

from bloc.util import conv_tf_matrix_to_json_compliant
from bloc.util import get_bloc_doc_lst
from bloc.util import get_bloc_variant_tf_matrix
from bloc.util import get_default_symbols

logger = logging.getLogger('bloc.bloc')

def run_subcommands(args, subcommand, payload):
    
    bloc_variant = None if args.ngram > 1 else {'type': 'folded_words', 'fold_start_count': args.fold_start_count, 'count_applies_to_all_char': False}
    if( args.token_pattern == 'bigram' ):
        token_pattern = '([^ |()*])'
    elif( args.token_pattern == 'word' ):
        token_pattern = '[^□⚀⚁⚂⚃⚄⚅. |()*]+|[□⚀⚁⚂⚃⚄⚅.]'
    else:
        token_pattern = args.token_pattern

    if( subcommand == 'top_ngrams' ):
        args.keep_tf_matrix = True
        args.set_top_ngrams = True
        args.top_ngrams_add_all_docs = True
        logger.info('\nSetting --keep-tf-matrix, --set-top-ngrams, and --top-ngrams-add-all-docs to True.');


    bloc_model = {
        'ngram': args.ngram,
        'bloc_variant': bloc_variant,
        'bloc_alphabets': args.bloc_alphabets,
        'token_pattern': token_pattern,
        'tf_matrix_norm': args.tf_matrix_norm,                  #set to '' if tf_matrices['tf_matrix_normalized'] not needed
        'keep_tf_matrix': args.keep_tf_matrix,
        'set_top_ngrams': args.set_top_ngrams,                  #set to False if tf_matrices['top_ngrams']['per_doc'] not needed. If True, keep_tf_matrix must be True
        'top_ngrams_add_all_docs': args.top_ngrams_add_all_docs #set to False if tf_matrices['top_ngrams']['all_docs'] not needed. If True, keep_tf_matrix must be True
    }

    all_bloc_symbols = get_default_symbols()
    bloc_collection = [ubloc for ubloc in payload]

    #generate collection of BLOC documents
    bloc_doc_lst = get_bloc_doc_lst(bloc_collection, bloc_model['bloc_alphabets'], src=args.account_src, src_class=args.account_class)
    
    tf_matrices = get_bloc_variant_tf_matrix(
        bloc_doc_lst, 
        min_df=2, 
        ngram=bloc_model['ngram'], 
        tf_matrix_norm=bloc_model['tf_matrix_norm'], 
        keep_tf_matrix=bloc_model['keep_tf_matrix'], 
        token_pattern=bloc_model['token_pattern'], 
        bloc_variant=bloc_model['bloc_variant'], 
        set_top_ngrams=bloc_model['set_top_ngrams'], 
        top_ngrams_add_all_docs=bloc_model['top_ngrams_add_all_docs']
    )

    if( subcommand == 'top_ngrams' ):
        report = print_top_ngrams(tf_matrices)
        report['users'] = [ u['screen_name'] for u in bloc_doc_lst ]
    elif( subcommand == 'sim' ):
        report = pairwise_usr_cmp(tf_matrices)
    
    return report

def feature_importance(vocab, fst_vect, sec_vect, k=10):

    feat_importance = []
    for i in range( len(vocab) ):
        feat = vocab[i]
        feat_importance.append({ 'feat': feat, 'score': fst_vect[i] * sec_vect[i] })
    feat_importance = sorted( feat_importance, key=lambda x: x['score'], reverse=True )[:k]
    
    counter = 1
    for f in feat_importance:

        logger.info( '\t{:>4} {:.4f} {}'.format(f'{counter}.', f['score'], f['feat']) )
        counter += 1

    return feat_importance

def pairwise_usr_cmp(tf_mat):

    logger.info('\npairwise_usr_cmp():')    

    if( 'tf_idf_matrix' not in tf_mat ):
        logger.info('tf_idf_matrix not in tf_mat, returning')
        return []
    
    report = []
    avg_sim = []
    tf_mat = conv_tf_matrix_to_json_compliant(tf_mat)
    indices = list(range(len(tf_mat['tf_idf_matrix'])))     
    pairs = combinations(indices, 2)

    for fst_u_indx, sec_u_indx in pairs:
        fst_u = tf_mat['tf_idf_matrix'][fst_u_indx]
        sec_u = tf_mat['tf_idf_matrix'][sec_u_indx]

        sim = cosine_similarity( [fst_u['tf_vector']], [sec_u['tf_vector']] )[0][0]
        avg_sim.append(sim)
        report.append({
            'sim': sim,
            'user_pair_indx': (fst_u_indx, sec_u_indx),
            'user_pair': (fst_u['screen_name'], sec_u['screen_name'])
        })

    report = sorted( report, key=lambda x: x['sim'], reverse=True )
    avg_sim = 0 if avg_sim == [] else sum(avg_sim)/len(avg_sim)

    logger.info('\nFeatures importance,')
    for i in range( len(report) ):
        
        r = report[i]
        logger.info( '\t{} vs. {}, (score, feature):'.format(r['user_pair'][0], r['user_pair'][1]) )
        
        fst_u_indx, sec_u_indx = r['user_pair_indx']
        r['feature_importance'] = feature_importance(tf_mat['vocab'], tf_mat['tf_idf_matrix'][fst_u_indx]['tf_vector'], tf_mat['tf_idf_matrix'][sec_u_indx]['tf_vector'])
        
        logger.info('')

    logger.info('Cosine sim,')
    for r in report:
        logger.info('\t{:.4f}: {} vs. {}'.format(r['sim'], r['user_pair'][0], r['user_pair'][1]))
        fst_u_indx, sec_u_indx = r['user_pair_indx']
    logger.info('\t------')
    logger.info('\t{:.4f}: Average cosine sim'.format(avg_sim))

    return report

def print_top_ngrams(tf_mat, k=10):

    def print_top_k_ngrams( top_ngrams, rate_key, skip_pause_glyph=True ):

        counter = 1
        for i in range( len(top_ngrams) ):

            n = top_ngrams[i]
            if( skip_pause_glyph is True and '□⚀⚁⚂⚃⚄⚅.'.find(n['term']) > -1 ):
                continue
            
            line = '\t{:<4} {:.4f} {} ({})'.format(f'{counter}.', n[rate_key], n['term'], get_word_type(n['term']) )
            counter += 1
            logger.info( line )

    logger.info('\nprint_top_ngrams():')

    if( 'top_ngrams' not in tf_mat ):
        logger.info('top_ngrams not in tf_mat, returning')
        return {}

    for i in range( len(tf_mat['top_ngrams']['per_doc']) ):
        
        topngrams = tf_mat['top_ngrams']['per_doc'][i]
        user = tf_mat['tf_idf_matrix'][i]['screen_name']        
        logger.info(f'\nTop {k} ngrams for user {user}, (term freq. TF, word):')
        print_top_k_ngrams( topngrams[:k], 'term_rate' )

    logger.info(f'\nTop {k} ngrams across all users, (document freq. DF, word):')
    print_top_k_ngrams( tf_mat['top_ngrams']['all_docs'][:k], 'doc_rate' )

    return tf_mat['top_ngrams']

    



import logging
import numpy as np
import sys

from copy import deepcopy
from numpy import linalg as LA
from scipy.spatial import distance
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score, recall_score, f1_score
from sklearn.metrics import classification_report
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from bloc.util import dumpJsonToFile
from bloc.util import genericErrorInfo
from bloc.util import get_bloc_variant_tf_matrix

from warnings import warn

logger = logging.getLogger('bloc.bloc')

def is_invalidate_matrix(matrix_key, tf_mat):
    if( matrix_key not in tf_mat or 'vocab' not in tf_mat ):
        logger.warning(f'\t{matrix_key} not in tf_mat, returning, see: ' + str(tf_mat.keys()) )
        return True

def get_bloc_tf_matrix(doc_lst, bloc_model):

    logger.info('\nget_bloc_tf_matrix()')
    logger.info('\tdoc_lst.len:', len(doc_lst))
    logger.info('\tngram:', bloc_model['ngram'])
    logger.info('\ttoken_pattern:', bloc_model['token_pattern'])
    logger.info('\tbloc_variant:', bloc_model['bloc_variant'])
    
    logger.info('\tget_bloc_variant_tf_matrix() - start')
    tf_mat = get_bloc_variant_tf_matrix(
        doc_lst,
        bloc_model['ngram'],
        token_pattern=bloc_model['token_pattern'],
        bloc_variant=bloc_model['bloc_variant'],
        add_all_docs=False
    )
    logger.info('\tget_bloc_variant_tf_matrix() - end')
    
    return tf_mat

def get_one_v_rest_centroid( dataset, vector_key, user ):

    count = 0
    centroid = []
    
    for u in dataset:
        
        if( user['user'] == u['user'] ):
            continue

        if( user['class'] != u['class'] ):
            continue

        if( len(centroid) == 0 ):
            centroid = np.zeros( len(u[vector_key]) )

        count += 1
        centroid += u[vector_key]

    if( count == 0 ):
        return {}

    return {
        'centroid': centroid/count,
        'count': count
    }

def predict_user_type_centroid_helper(user_vect, candidates, dist_metric='cosine'):

    min_dist = {'class': '', 'dist': sys.maxsize}
    for src, clss_dct in candidates.items():
        
        if( dist_metric == 'cosine' ):
            if( isinstance(user_vect, csr_matrix) ):
                user_vect = user_vect.toarray()[0]
            dist = distance.cosine(user_vect, clss_dct['centroid'])
        else:
            dist = LA.norm(user_vect - clss_dct['centroid'])
        

        if( dist < min_dist['dist'] ):
            min_dist['class'] = clss_dct['class']
            min_dist['dist'] = dist
            min_dist['src'] = src
    
    return min_dist['class']

def run_pca(matrix, features, n_components=10, z_score_normalize=True):

    n_features = len(features)
    if( n_features < 3 ):
        return {}
    
    if( n_components > n_features ):
        warn(f'run_pca(): setting n_components = 2 since n_components ({n_components}) > n_features ({n_features})')
        n_components = 2

    
    pca = PCA(n_components=n_components) # estimate only 2 PCs

    #X = np.array( [[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]] )
    X = [ v['tf_vector'] for v in matrix ]
    
    # Z-score the features
    #this might not be needed since tf-idf already performs some level of normalization
    if( z_score_normalize is True ):
        scaler = StandardScaler()
        scaler.fit(X)
        X = scaler.transform(X)
    

    # The PCA model
    #credit: https://medium.com/ai-in-plain-english/how-to-implement-pca-with-python-and-scikit-learn-22f3de4e5983
    #for model training/testing see: https://stackabuse.com/implementing-pca-in-python-with-scikit-learn/ and https://towardsdatascience.com/pca-using-python-scikit-learn-e653f8989e60
    #pca = PCA(n_components=n_components) # estimate only 2 PCs
    X_new = pca.fit_transform(X) # project the original data into the PCA space

    #The PCA class contains explained_variance_ratio_ which returns the variance caused by each of the principal components. 
    explained_variance = pca.explained_variance_ratio_
    explained_variance_sum = sum(explained_variance)
    abs_eigenvector = abs(pca.components_)

    if( n_components > len(explained_variance) ):
        n_components = len(explained_variance)
    
    feature_importance = []
    for i in range( len(features) ):

        score = 0
        for j in range(n_components):
            score += abs_eigenvector[:,i][j] * explained_variance[j]

        score = score/explained_variance_sum
        #score is weighted average of eigenvector components (where weights are explained variance ratio of PC)
        feature_importance.append({
            'feature': features[i],
            'score': score,
            'indx': i
        })

    feature_importance = sorted( feature_importance, key=lambda x: x['score'], reverse=True )

    return {
        'pca': pca,
        'pca_points': [ {'pca_vect': v} for v in X_new ],
        'feature_importance': feature_importance
    }

def get_bloc_pred_centroids( tf_mat, vocab, n_components=0, **kwargs ):

    if( len(tf_mat) == 0 or len(vocab) == 0 ):
        return {}
    
    pca_payload = {}
    centroids = {}
    
    kwargs.setdefault('pca_z_score_normalize', False)

    if( n_components > 0 ):
        
        try:
            pca_payload = run_pca( tf_mat, vocab, n_components=n_components, z_score_normalize=kwargs['pca_z_score_normalize'] )
        except:
            genericErrorInfo()

        if( len(pca_payload) == 0 ):
            return {}

        n_components = len(pca_payload['pca_points'][0]['pca_vect'])#it's possible for n_components to change after run_pca
        centroid_vect_dim = n_components
    else:
        centroid_vect_dim = len( tf_mat[0]['tf_vector'] )


    for i in range( len(tf_mat) ):
        
        user = tf_mat[i]
        classname = user['class']
        centroids.setdefault(classname, {'centroid': np.zeros(centroid_vect_dim), 'count': 0, 'class': classname})
        centroids[classname]['count'] += 1

        if( n_components > 0 and len(pca_payload) > 0 ):
            pca_payload['pca_points'][i]['user'] = user['user']
            pca_payload['pca_points'][i]['class'] = classname

            #centroid is calculated from PCA
            centroids[classname]['centroid'] += pca_payload['pca_points'][i]['pca_vect']
        else:
            #centroid is calculated from raw tf_vector
            centroids[classname]['centroid'] += tf_mat[i]['tf_vector']


    for classname in centroids:
        centroids[classname]['centroid'] = centroids[classname]['centroid']/centroids[classname]['count']

    return {
        'pca_payload': pca_payload,
        'centroids': centroids
    }

def predict_user_class_centroid(doc_lst, tf_mat, centroids, pca_payload, one_v_rest=True):

    if( len(doc_lst) == 0 or len(tf_mat) == 0 ):
        return {}

    all_classes = list(centroids.keys())
    all_classes.sort()

    y_true = []
    y_pred = []
    class_rep = {}
    conf_mat = np.zeros( (len(all_classes), len(all_classes)) )
    
    for i in range( len(doc_lst) ):

        user = doc_lst[i]
        classname = user['class']
        loc_centroid = centroids if one_v_rest is False else deepcopy(centroids)

        if( len(pca_payload) == 0 ):
            #train_split = -1: suggests doc_lst is a test set and tf_mat was generated with test set, but centroids was calculated with training set, so no need to replace centroid that includes user with get_one_v_rest_centroid()
            #empty pca_payload means centroids are calculated from raw TF matrices.
            #get_one_v_rest_centroid recalculates centroid, but ensures user is not part of te calculation
            
            loc_centroid[classname] = loc_centroid[classname] if one_v_rest is False else get_one_v_rest_centroid( tf_mat, 'tf_vector', user )
            pred = predict_user_type_centroid_helper( tf_mat[i]['tf_vector'], loc_centroid )
        else:
            
            loc_centroid[classname] = loc_centroid[classname] if one_v_rest is False else get_one_v_rest_centroid( pca_payload['pca_points'], 'pca_vect', user )
            pred = predict_user_type_centroid_helper( pca_payload['pca_points'][i]['pca_vect'], loc_centroid )
        
        
        pred_i = all_classes.index(pred)
        actu_i = all_classes.index(user['class'])
        conf_mat[pred_i, actu_i] += 1

        y_true.append(user['class'])
        y_pred.append(pred)
        
        '''
        print( i, 'user:', user['user'] )
        print( 'Pred/Actu:', pred, user['class'] )
        print(conf_mat)
        print()
        '''

    if( len(y_true) != 0 and len(y_pred) != 0 ):
        class_rep = classification_report(y_true, y_pred, output_dict=True, target_names=all_classes)
    
    return {
        'confusion_matrix': conf_mat,
        'all_classes': all_classes,
        'classification_report': class_rep
    }

def predict_user_class_tfidf(training_doc_lst, test_doc_lst, model, **kwargs):

    if( len(training_doc_lst) == 0 or len(test_doc_lst) == 0 ):
        logger.warning('len(training_doc_lst) == 0 or len(test_doc_lst) == 0, returning')
        return {}

    kwargs.setdefault('matrix_key', 'tf_idf_matrix')
    min_df = kwargs.get('min_df', 1)

    matrix_key = kwargs['matrix_key']

    train_per_src_doc = {}
    details = {}
    all_train_classes = set()
    all_test_classes = set([ d['class'] for d in test_doc_lst ])

    
    for i in range( len(training_doc_lst) ):
        src = training_doc_lst[i]['src']
        classs = training_doc_lst[i]['class']
        cent_src = f'{src}.{classs}'

        train_per_src_doc.setdefault( cent_src, {'user_indices': [], 'class': classs} )
        train_per_src_doc[cent_src]['user_indices'].append(i)
        all_train_classes.add( classs )


    if( len(all_train_classes - all_test_classes) > 0 ):
        #all_test_classes MUST be proper subset of all_train_classes
        logger.warning(' all_test_classes has class(es) absent from all_train_classes, returning')
        return {}

    
    train_test_tf_mat = get_bloc_variant_tf_matrix(
        training_doc_lst + test_doc_lst,
        model['ngram'],
        token_pattern=model['token_pattern'],
        bloc_variant=model['bloc_variant'],
        min_df=min_df,
        add_all_docs=False
    )
    
    if( is_invalidate_matrix( matrix_key, train_test_tf_mat ) ):
        logger.warning(' is_invalidate_matrix( matrix_key, train_test_tf_mat ), returning')
        return {}

    if( len(train_test_tf_mat[matrix_key]) == 0 ):
        logger.warning(' train_test_tf_mat is empty, returning')
        return {}

    details['vocab'] = train_test_tf_mat['vocab']
    #get centroids for training models - start
    for src in train_per_src_doc:
        cent = np.zeros( len(details['vocab']) )
        for indx in train_per_src_doc[src]['user_indices']:
            cent += train_test_tf_mat[matrix_key][indx]['tf_vector']
        
        cent = cent / len(train_per_src_doc[src]['user_indices'])
        train_per_src_doc[src]['centroid'] = cent
    #get centroids for training models - end

    all_train_classes = list(all_train_classes)
    all_train_classes.sort()

    y_true = []
    y_pred = []
    class_rep = {}
    conf_mat = np.zeros( (len(all_train_classes), len(all_train_classes)) )

    for i in range( len(test_doc_lst) ):

        u = test_doc_lst[i]
        classname = u['class']
        test_tf_mat_indx = len(training_doc_lst) + i#because training_doc_lst + test_doc_lst

        if( u['user_id'] != train_test_tf_mat[matrix_key][test_tf_mat_indx]['user_id'] ):
            logger.warning('\npredict_user_class_tfidf(): UNEXPECTED STATE, USER MISMATCH\n')
            continue

        pred = predict_user_type_centroid_helper( train_test_tf_mat[matrix_key][test_tf_mat_indx]['tf_vector'], train_per_src_doc, dist_metric=kwargs.get('dist_metric', 'cosine') )
        
        pred_i = all_train_classes.index(pred)
        actu_i = all_train_classes.index(u['class'])
        conf_mat[pred_i, actu_i] += 1

        y_true.append(u['class'])
        y_pred.append(pred)
        
        '''
        print( i, 'user:', u['user'] )
        print( 'Pred/Actu:', pred, u['class'] )
        print(conf_mat)
        print()
        '''

    if( len(y_true) != 0 and len(y_pred) != 0 ):
        class_rep = classification_report(y_true, y_pred, output_dict=True, target_names=all_train_classes)
    
    return {
        'confusion_matrix': conf_mat,
        'all_classes': all_train_classes,
        'details': details,
        'classification_report': class_rep
    }

def predict_user_class_bloc_centroid(bloc_model, n_components=0, training_doc_lst=[], test_doc_lst=[], **kwargs):

    '''
        n_components: if < 1 means no PCA vector should be used, else PCA would be used to calculate centroid
    
        if only training_doc_lst.len > 0, model is trained with entire training_doc_lst and evaluated with one v rest method
        if training_doc_lst.len > 0 and test_doc_lst.len > 0, model is trained with training_doc_lst and tested on test_doc_lst
    '''
    if( bloc_model not in ['bigram', 'word'] ):
        logger.warning(f'{bloc_model} not bigram or word, returning')
        return {}

    bloc_model_params = {
        'bigram': {
                'ngram': 2,
                'token_pattern': '[^ |()*]',
                'bloc_variant': None
            },
        'word': {
                'ngram': 1,
                'token_pattern': '[^□⚀⚁⚂⚃⚄⚅ |()*]+|[□⚀⚁⚂⚃⚄⚅]',
                'bloc_variant': {'type': 'folded_words', 'fold_start_count': 4, 'count_applies_to_all_char': False}
            }
        }

    kwargs.setdefault('matrix_key', 'tf_idf_matrix')
    matrix_key = kwargs['matrix_key']
    #train_split=-1

    if( len(training_doc_lst) > 0 and len(test_doc_lst) > 0 ):

        training_tf_mat = get_bloc_tf_matrix( training_doc_lst, bloc_model_params[bloc_model] )
        if( is_invalidate_matrix( matrix_key, training_tf_mat ) ):
            return {}

        test_tf_mat = get_bloc_tf_matrix( test_doc_lst, bloc_model_params[bloc_model] )
        if( is_invalidate_matrix( matrix_key, test_tf_mat ) ):
            return {}

        training_cent_pca = get_bloc_pred_centroids( training_tf_mat[matrix_key], training_tf_mat['vocab'], n_components=n_components, **kwargs )
        test_cent_pca = get_bloc_pred_centroids( test_tf_mat[matrix_key], test_tf_mat['vocab'], n_components=n_components, **kwargs )
        if( len(training_cent_pca) == 0 or len(test_cent_pca) == 0 ):
            return {}
        class_rep = predict_user_class_centroid( test_doc_lst, test_tf_mat[matrix_key], training_cent_pca['centroids'], test_cent_pca['pca_payload'], one_v_rest=False )

    else:

        tf_mat = get_bloc_tf_matrix( training_doc_lst, bloc_model_params[bloc_model] )
        if( is_invalidate_matrix( matrix_key, tf_mat ) ):
            return {}
        
        cent_pca = get_bloc_pred_centroids( tf_mat[matrix_key], tf_mat['vocab'], n_components=n_components, **kwargs )
        class_rep = predict_user_class_centroid( training_doc_lst, tf_mat[matrix_key], cent_pca['centroids'], cent_pca['pca_payload'], one_v_rest=True )

    return class_rep

def predict_user_class_bloc_tfidf(bloc_model, training_doc_lst=[], test_doc_lst=[], **kwargs):

    if( bloc_model not in ['bigram', 'word'] ):
        logger.warning(f'{bloc_model} not bigram or word, returning')
        return {}

    if( len(training_doc_lst) == 0 or len(test_doc_lst) == 0 ):
        return {}

    parenth_flag = '' if kwargs.get('omit_content_parentheses', False) is True else '()'
    pause_flag = '|[□⚀⚁⚂⚃⚄⚅]' if kwargs.get('add_pauses', True) is True else ''
    word_token_pattern = f'[^□⚀⚁⚂⚃⚄⚅ |*{parenth_flag}]+{pause_flag}'

    #print('\tword_token_pattern:', word_token_pattern)

    default_model_params = {
        'bigram': {
                'ngram': 2,
                'token_pattern': '[^ |()*]',
                'bloc_variant': None
            },
        'word': {
                'ngram': 1,
                'token_pattern': word_token_pattern,
                'bloc_variant': {'type': 'folded_words', 'fold_start_count': 4, 'count_applies_to_all_char': False}
        }
    }
    
    bloc_model_params = kwargs.get('bloc_model_params', default_model_params[bloc_model])

    if( len(bloc_model_params) == 0 ):
        print('len(bloc_model_params) == 0, returning')
        return {}

    class_rep = predict_user_class_tfidf( training_doc_lst, test_doc_lst, bloc_model_params, **kwargs )
        
    return class_rep


import gzip

from bloc.generator import add_bloc_sequences
from bloc.util import genericErrorInfo
from bloc.util import getDictFromJson

def user_tweets_generator_0(filenames, pos_id_mapping, gen_bloc_params, **kwargs):
    
    '''
        Expects gzip compressed file.
        File format (user_id_k is optional):
        user_id_0\tJSON list of tweets
        user_id_1\tJSON list of tweets
        user_id_2\tJSON list of tweets

    '''

    index = -1
    rm_doc_text = kwargs.get('rm_doc_text', True)
    keep_bloc_details = kwargs.get('keep_bloc_details', ['created_at_utc', 'screen_name', 'user_id'])

    for f in filenames:
        with gzip.open(f, 'rb') as file:
            for row in file:
                
                doc = ''
                index += 1
                
                try:
                    row = row.decode().split('\t')
                    if( len(row) == 0 ):
                        continue
                    
                    tweets = row[-1]
                    tweets = getDictFromJson(tweets)
                    bloc_payload = add_bloc_sequences(tweets, **gen_bloc_params)

                    if( len(tweets) != 0 and 'bloc' in bloc_payload ):
                        doc = [ bloc_payload['bloc'][dim] for dim in bloc_payload['bloc'] ]
                        doc = ' '.join(doc).strip()
                except:
                    genericErrorInfo()

                if( isinstance(pos_id_mapping, dict) ):
                    
                    pos_id_mapping[index] = {'id': index}

                    for ky in keep_bloc_details:
                        if( ky in bloc_payload ):
                            pos_id_mapping[index][ky] = bloc_payload[ky]

                    if( rm_doc_text is False ):
                        pos_id_mapping[index]['text'] = doc

                yield doc


from datetime import datetime
from copy import deepcopy
def get_v1_user_obj(v2_users):

    all_users = {}
    for u in v2_users:

        u['screen_name'] = u.pop('username', '')
        u['followers_count']  = u['public_metrics'].pop('followers_count', -1)
        u['friends_count']  = u['public_metrics'].pop('following_count', -1)
        u['statuses_count']  = u['public_metrics'].pop('tweet_count', -1)
        u['listed_count']  = u['public_metrics'].pop('listed_count', -1)
        u['favourites_count'] = -1
        u['created_at'] = datetime.strptime(u['created_at'], '%Y-%m-%dT%H:%M:%S.000Z').strftime('%a %b %d %H:%M:%S +0000 %Y')

        all_users[ u['id'] ] = u

    return all_users

def get_referenced_v2_tweets(v2_tweets):
    
    all_twts = {}
    for t in v2_tweets:
        all_twts[ t['id'] ] = t

    return all_twts

def get_v1_lite_twt(v2_tweet, all_v1_user_objs, all_referenced_v2_tweets):
    
    def get_in_reply_to_status_id(v2_tweet):

        for t in v2_tweet.get('referenced_tweets', []):
            if( t['type'] == 'replied_to' ):
                return t['id']

        return None

    def get_v1_entities(v2_entities):

        if( 'mentions' in v2_entities ):
            for i in range( len(v2_entities['mentions']) ):

                m = v2_entities['mentions'][i]
                m['screen_name'] = m.pop('username', '')
                m['indices'] = [ m.pop('start', -1), m.pop('end', -1) ]

            v2_entities['user_mentions'] = v2_entities.pop('mentions', [])

        if( 'hashtags' in v2_entities ):
            for i in range( len(v2_entities['hashtags']) ):
                h = v2_entities['hashtags'][i]
                h['text'] = h.pop('tag', '')
                h['indices'] = [ h.pop('start', -1), h.pop('end', -1) ]

        if( 'cashtags' in v2_entities ):
            for i in range( len(v2_entities['cashtags']) ):
                c = v2_entities['cashtags'][i]
                c['text'] = c.pop('tag', '')
                c['indices'] = [ c.pop('start', -1), c.pop('end', -1) ]
            v2_entities['symbols'] = v2_entities.pop('cashtags', [])

        if( 'urls' in v2_entities ):
            for i in range( len(v2_entities['urls']) ):
                u = v2_entities['urls'][i]
                u['indices'] = [ u.pop('start', -1), u.pop('end', -1) ]


        v2_entities.setdefault('user_mentions', [])
        v2_entities.setdefault('hashtags', [])
        v2_entities.setdefault('symbols', [])
        v2_entities.setdefault('urls', [])
        return v2_entities


    if( 'id' not in v2_tweet ):
        return {}

    if( v2_tweet['author_id'] not in all_v1_user_objs ):
        return {}


    in_reply_to_user_id = v2_tweet.get('in_reply_to_user_id', None)
    in_reply_to_screen_name = None if in_reply_to_user_id is None else all_v1_user_objs.get(in_reply_to_user_id, {}).get('screen_name', None)
    place = v2_tweet.get('geo', {}).get('place_id', None) #assumption is that geo, coordinates and places have collapsed into place
    payload = {
        'id': v2_tweet['id'],
        'full_text': v2_tweet.get('text', ''),
        'source': v2_tweet.get('source', ''),
        'lang': v2_tweet.get('lang', ''),
        'user': deepcopy(all_v1_user_objs[ v2_tweet['author_id'] ]),
        'geo': None,
        'coordinates': None,
        'place': place if place is None else {'id': place},
        'in_reply_to_status_id': get_in_reply_to_status_id(v2_tweet),
        'in_reply_to_screen_name': in_reply_to_screen_name,
        'in_reply_to_user_id': in_reply_to_user_id,
        'created_at': datetime.strptime(v2_tweet['created_at'], '%Y-%m-%dT%H:%M:%S.000Z').strftime('%a %b %d %H:%M:%S +0000 %Y'),
        'entities': get_v1_entities(deepcopy( v2_tweet.get('entities', {}) ))
    }

    if( 'context_annotations' in v2_tweet ):
        payload['context_annotations'] = v2_tweet['context_annotations']


    #add retweeted_status - start
    r_twts = v2_tweet.get('referenced_tweets', [])
    if( len(r_twts) == 1 ):

        twtype = r_twts[0]['type']; rtid = r_twts[0]['id']
        if( twtype == 'retweeted' and rtid in all_referenced_v2_tweets ):
            payload['retweeted_status'] = get_v1_lite_twt( all_referenced_v2_tweets[rtid], all_v1_user_objs, all_referenced_v2_tweets ) 
    #add retweeted_status - end


    #add extended_entities - start
    if( 'attachments' in v2_tweet and 'media_keys' in v2_tweet['attachments'] ):
        media = [ m.split('_')[-1] for m in v2_tweet['attachments']['media_keys'] ]
        media = [ {'id': m} for m in media ]
        payload['extended_entities'] = {'media': media}
    #add extended_entities - end
    
    return payload

def conv_v2_tweets_to_v1(v2_tweets):

    if( 'data' not in v2_tweets ):
        return []

    all_v1_user_objs = get_v1_user_obj( v2_tweets.get('includes', {}).get('users', []) )
    all_referenced_v2_tweets = get_referenced_v2_tweets( v2_tweets.get('includes', {}).get('tweets', []) )
    
    return [ get_v1_lite_twt(t, all_v1_user_objs, all_referenced_v2_tweets) for t in v2_tweets['data'] ]


import json
import logging
import os, sys
import osometweet
import re
import time

from copy import deepcopy
from datetime import datetime, timedelta
from requests_oauthlib import OAuth1Session
from textblob import TextBlob

from bloc.util import color_bloc_action_str
from bloc.util import datetimeFromUtcToLocal
#from bloc.util import dumpJsonToFile
from bloc.util import find_tweet_timestamp_post_snowflake
from bloc.util import genericErrorInfo
from bloc.util import getDictFromFile
from bloc.util import getDictFromJsonGZ
from bloc.util import get_screen_name_frm_status_uri
from bloc.util import gen_post_snowflake_twitter_id
from bloc.util import gzipTextFile
from bloc.util import twitter_v2_user_lookup_ids
from bloc.v2_support import conv_v2_tweets_to_v1

logger = logging.getLogger('bloc.bloc')
'''
    Alphabets
    Action
    Content-Syntactic
    Content-Semantic-Sentiment
    Change
    Time

    BLOC Master Alphabets list
    - - Content-Semantic-Sentiment    (- - Neutral)
    a - Change                        (Profile appearance change)
    D - Change                        (Delete tweet)
    d - Change                        (Description change)
    E - Content-Syntactic             (E - Media)
    F - Change                        (Follow someone)
    f - Change                        (Unfollow someone)
    g - Change                        (Profile location change)
    G - Change                        (geographical location change (see coordinates and place))
    H - Content-Syntactic             (H - Hashtag)
    L - Change                        (Liked tweet)
    l - Change                        (Unliked tweet)
    M - Content-Syntactic             (M - Mention of friend (can't be checked until friendship relationship assigned))
    m - Content-Syntactic             (m - Mention of non-friend)
    n - Change                        (Name change)
    N - Change                        (Handle change)
    P - Action                        (Reply a friend (can't be checked until friendship relationship assigned))
    p - Action                        (Reply a non-friend)
    q - Content-Syntactic             (q - Quote URL)
    R - Action                        (Retweet a friend (can't be checked until friendship relationship assigned))
    r - Action                        (Retweet a non-friend)
    s - Change                        (source change)
    T - Action                        (Tweet)
    t - Content-Syntactic             (t - Text)
    u - Change                        (URL change)
    U - Content-Syntactic             (U - URL)
    W - Change                        (Gained followers)
    w - Change                        (Lossed followers)
    x - Content-Semantic-Entities     (Product (e.g., Mountain Dew, Mozilla Firefox))
    ¤ - Content-Syntactic             (¤ - Cashtag)
    λ - Change                        (Language change)
    π - Action                        (Reply self)
    ρ - Action                        (Retweet self)
    φ - Content-Syntactic             (φ - Quote self)
    ⊛ - Content-Semantic-Entities     (Other (e.g., Diabetes, Super Bowl 50))
    ⋂ - Content-Semantic-Sentiment    (⋂ - Negative)
    ⋃ - Content-Semantic-Sentiment    (⋃ - Positive)
    ⋈ - Content-Semantic-Entities     (Organization (e.g., Chicago White Sox, IBM))
    ⌖ - Content-Semantic-Entities     (Place (e.g., Detroit, Cali, or "San Francisco, California"))
    □ - Time                          (under minute_mark)
    ⚀ - Time                          (under hour)
    ⚁ - Time                          (under day)
    ⚂ - Time                          (under week)
    ⚃ - Time                          (under month)
    ⚄ - Time                          (under year)
    ⚅ - Time                          (over year)
    ⚇ - Content-Semantic-Entities     (Person (e.g., Barack Obama, Daniel, or George W. Bush))
'''

def friendship_lookup(oauth, source_screen_name, target_screen_name, msg=''):

    #1 request per 5 seconds: https://developer.twitter.com/en/docs/twitter-api/v1/accounts-and-users/follow-search-get-users/api-reference/get-friendships-show
    logger.info('friendship_lookup(): sleeping for 5 seconds' + msg)
    time.sleep(5)

    params = {
        'source_screen_name': source_screen_name,
        'target_screen_name': target_screen_name
    }
    try:
        response = oauth.get('https://api.twitter.com/1.1/friendships/show.json', params=params)
        friendship = json.loads(response.text)['relationship']

        return friendship
    except:
        genericErrorInfo()

    return {}

def fmt_twt_time_to_loc(created_at):
    created_at = datetime.strptime(created_at, '%a %b %d %H:%M:%S %z %Y')

    local_time = datetimeFromUtcToLocal( created_at )
    local_time = datetime.strftime( local_time, '%Y-%m-%d %H:%M:%S' )   

    return local_time

def get_tweet_type(twt):

    if( 'in_reply_to_status_id' not in twt ):
        return ''

    if( twt['in_reply_to_status_id'] is not None ):
        return 'reply'
    elif( 'retweeted_status' in twt ):
        return 'retweet'
    else:
        return 'tweet'

def get_timeline_tweets(oauth, screen_name, user_id='', max_pages=1, following_lookup=True, timeline_startdate='', timeline_scroll_by_hours=None):

    if( max_pages < 0 ):
        return []


    params = {'tweet_mode': 'extended', 'count': 20}
    screen_name = screen_name.strip()
    user_id = user_id.strip()
    
    if( screen_name == '' and user_id == '' ):
        return []

    if( screen_name != '' ):
        params['screen_name'] = screen_name
    elif( user_id != '' ):
        params['user_id'] = user_id


    tweets = []
    dedup_set = set()
    timeline_startdate = '' if timeline_startdate == '' else datetime.strptime( timeline_startdate, '%Y-%m-%d %H:%M:%S' )

    if( isinstance(timeline_scroll_by_hours, int) is False ):
        timeline_startdate = ''


    logger.info('\nget_timeline_tweets():')
    for i in range(max_pages):

        user_id_det = '' if user_id == '' else f' (id: {user_id})'
        logger.info('\tpage: ' + str(i+1) + ' of ' + str(max_pages) + ', for: @' + screen_name + f'{user_id_det}')

        #guess max_id, since_id - start
        if( timeline_startdate != '' ):
            try:
                timeline_next_date = timeline_startdate + timedelta(hours=timeline_scroll_by_hours)
                next_twt_id = gen_post_snowflake_twitter_id(timeline_startdate)

                params.pop('max_id', None)
                params.pop('since_id', None)

                if( timeline_scroll_by_hours > 0 ):
                    params['since_id'] = next_twt_id
                else:
                    params['max_id'] = next_twt_id

                logger.info('\ttimeline_startdate: ' + str(timeline_startdate) + ' (' + str(timeline_scroll_by_hours) + ') -> ' + str(next_twt_id) )

            except:
                genericErrorInfo()
        #guess max_id, since_id - end


        try:
            response = oauth.get("https://api.twitter.com/1.1/statuses/user_timeline.json", params=params)
            timeline = json.loads(response.text)
        except:
            genericErrorInfo()
            

        if( 'errors' in timeline ):
            logger.error( '\terror: ' + str(timeline['errors']) )
            #to conform with 1 request per second rate limit for user auth: https://developer.twitter.com/en/docs/twitter-api/v1/tweets/timelines/api-reference/get-statuses-user_timeline
            time.sleep(1)
            continue

        if( isinstance(timeline, dict) or len(timeline) == 0 ):
            #to conform with 1 request per second rate limit for user auth: https://developer.twitter.com/en/docs/twitter-api/v1/tweets/timelines/api-reference/get-statuses-user_timeline
            time.sleep(1)
            continue
        
        
        #dedup tweets - start
        for twt in timeline:
            
            if( 'id' not in twt ):
                continue

            if( twt['id'] in dedup_set ):
                continue

            dedup_set.add( twt['id'] )
            tweets.append(twt)
        #dedup tweets - end
        logger.info( f'\t{len(tweets)} tweets' )


        if( timeline_startdate == '' ):
            #use default timeline pagination since user is not paginating by time
            params['max_id'] = timeline[-1]['id'] - 1
        else:
            timeline_startdate = timeline_next_date
            logger.info('\tpaginating with datetime: ' + str(timeline_startdate) + '\n' )


        #add bloc key to tweets
        for j in range( len(timeline) ):
            timeline[j].setdefault('bloc', {
                'src_follows_tgt': None,
                'tgt_follows_src': None
            })

        sleep_flag = True
        if( following_lookup is True ):
            #no need to sleep for 1 second since friendship look up sleeps for 5 seconds.
            
            timeline_len = len(timeline)
            for j in range( timeline_len ):
                
                twt = timeline[j]
                if( 'in_reply_to_status_id' not in twt ):
                    continue

                tweet_type = get_tweet_type(twt)
                if( tweet_type not in ['reply', 'retweet'] ):
                    continue

                if( tweet_type == 'reply' ):
                    tgt = twt['in_reply_to_screen_name']
                elif( tweet_type == 'retweet' ):
                    tgt = twt['retweeted_status']['user']['screen_name']
                else:
                    continue

                #skip self
                if( screen_name == tgt ):
                    continue
                
                msg = ', for ' + tweet_type + ', ' + str(j+1) + ' of ' + str(timeline_len)
                relationship = friendship_lookup(oauth, screen_name, tgt, msg=msg)
                sleep_flag = False

                if( len(relationship) != 0 ):
                    twt['bloc']['src_follows_tgt'] = relationship['source']['following']
                    twt['bloc']['tgt_follows_src'] = relationship['source']['followed_by']
                    
        if( sleep_flag ):
            #to conform with 1 request per second rate limit for user auth: https://developer.twitter.com/en/docs/twitter-api/v1/tweets/timelines/api-reference/get-statuses-user_timeline
            time.sleep(1)

    return tweets

def v2_get_timeline_tweets(ostwt, user_id, max_pages=1, max_results=100, timeline_startdate='', timeline_scroll_by_hours=None):

    if( max_pages < 0 ):
        return []

    user_id = user_id.strip()
    if( user_id == '' ):
        return []


    tweets = []
    dedup_set = set()
    user_screen_name = ''
    next_token = None
    timeline_scroll_by_hours = 0 if timeline_scroll_by_hours is None else timeline_scroll_by_hours    
    timeline_startdate = '' if timeline_startdate == '' else datetime.strptime( timeline_startdate, '%Y-%m-%d %H:%M:%S' )

    logger.info('\nv2_get_timeline_tweets():')
    for i in range(max_pages):

        logger.info( f'\tpage: {i+1} of {max_pages} for id: {user_id}{user_screen_name}' )
        if( timeline_startdate != '' ):
            logger.info( f'\tfrom: {timeline_startdate} (timeline_scroll_by_hours: {timeline_scroll_by_hours})' )

        try:
            end_time = None if timeline_startdate == '' else timeline_startdate.strftime('%Y-%m-%dT%H:%M:%S') + 'Z' 
            response = ostwt.get_tweet_timeline(user_id, max_results=max_results, everything=True, pagination_token=next_token, end_time=end_time)
        except:
            genericErrorInfo()

        if( max_pages > 1 ):
            logger.info('\tsleeping for 3 seconds')
            time.sleep(3)

        if( 'meta' not in response or 'data' not in response ):
            logger.info('\tno meta and data keys, breaking')
            break        
        
        v1_tweets = conv_v2_tweets_to_v1(response)
        #dedup tweets - start
        for twt in v1_tweets:
            
            if( 'id' not in twt ):
                continue

            if( twt['id'] in dedup_set ):
                continue

            dedup_set.add( twt['id'] )
            user_screen_name = ' (@' + twt['user']['screen_name'] + ')'
            twt.setdefault('bloc', {'src_follows_tgt': None, 'tgt_follows_src': None})
            tweets.append(twt)
        #dedup tweets - end
        logger.info( '\t{} tweets, {} errors'.format(len(tweets), len(response.get('errors', []))) )


        next_token = response['meta'].get('next_token', None)
        #paginate - start
        if( timeline_startdate != '' and timeline_scroll_by_hours != 0 ):
            timeline_startdate = datetime.strptime( response['data'][-1]['created_at'], '%Y-%m-%dT%H:%M:%S.000Z' )
            timeline_startdate = timeline_startdate + timedelta(hours=timeline_scroll_by_hours)
            next_token = None
            logger.info('\ttime pagination, setting next_token to None')#time pagination and next_token are mutually exclusive

        elif( next_token is None ):
            logger.info('\tno next_token, breaking')
            break
        #paginate - end

        logger.info('')

    return tweets

def get_twt_text_exclusively(txt, entities):

    for ky, val in entities.items():
        for ent in val:
            
            if( 'indices' not in ent ):
                continue

            st, en = ent['indices']
            txt = txt[:st] + ' '*(en - st) + txt[en:]

    return txt

def get_action_glyphs():
    return ['P', 'p', 'π', 'R', 'r', 'ρ', 'T']

def get_word_type(word):
    
    bloc_types = {
        'action': set(get_action_glyphs()),
        'content_syntactic': set('EHMmqφtU'),
        'time': set('□⚀⚁⚂⚃⚄⚅.'),
        'change': set('aDdFfgGλLlnNsuWw'),
        'content_semantic_entities': set('⚇⌖⋈x⊛'),
        'content_semantic_sentiment': set('⋃-⋂')
    }
    
    word = set(word)
    for typ, tokens in bloc_types.items():
        if( len(word & tokens) > 0 ):
            return typ

    return ''

def get_delta_glyph(symbols, delta_seconds, blank_mark=60, minute_mark=5):
    
    '''
        Default time symbols
        □ under minute_mark
        ⚀ under hour
        ⚁ under day
        ⚂ under week
        ⚃ under month
        ⚄ under year
        ⚅ over year
    '''

    if( delta_seconds < blank_mark ):
        glyph = symbols['blank_mark']['symbol']
    #elif( delta_seconds < short_pause_mark ):
    #    glyph = '.' * ( (delta_seconds//60) )
    elif( delta_seconds < minute_mark*60 ):
        glyph = symbols['under_minute_mark']['symbol']
    elif( delta_seconds < 3600 ):
        glyph = symbols['under_hour_mark']['symbol']
    elif( delta_seconds < 86400 ):
        glyph = symbols['under_day_mark']['symbol']
    elif( delta_seconds < 604800 ):
        glyph = symbols['under_week_mark']['symbol']
    elif( delta_seconds < 2628000 ):
        glyph = symbols['under_month_mark']['symbol']
    elif( delta_seconds < 31540000 ):
        glyph = symbols['under_year_mark']['symbol']
    else:
        glyph = symbols['over_year_mark']['symbol']

    return glyph

def get_duration_prefix(symbols, cur_time, prev_time, blank_mark, minute_mark):

    seq_diff_time = datetime.strptime( cur_time, '%Y-%m-%d %H:%M:%S' ) - datetime.strptime( prev_time, '%Y-%m-%d %H:%M:%S' )
    #delta_seconds = (seq_diff_time.days*86400) + seq_diff_time.seconds
    delta_seconds = int( seq_diff_time.total_seconds() )

    dur_glyph = get_delta_glyph( symbols, delta_seconds, blank_mark=blank_mark, minute_mark=minute_mark )
    
    return delta_seconds, dur_glyph

def get_pause(symbols, twt, prev_twt, blank_mark, minute_mark, use_src_ref_time=False):

    if( 'in_reply_to_status_id' not in twt ):
        return -1, ''

    source_ref_time = ''
    delta_seconds = -1
    dur_glyph = ''

    if( use_src_ref_time is True ):
        if( twt['in_reply_to_status_id'] is not None ):#reply
            source_ref_time = find_tweet_timestamp_post_snowflake( twt['in_reply_to_status_id'] )
            source_ref_time = datetimeFromUtcToLocal( source_ref_time )
            source_ref_time = datetime.strftime( source_ref_time, '%Y-%m-%d %H:%M:%S' )
            
        elif( 'retweeted_status' in twt ):#retweet
            source_ref_time = twt['retweeted_status']['created_at']
            source_ref_time = datetime.strptime(source_ref_time, '%a %b %d %H:%M:%S %z %Y')
            source_ref_time = datetimeFromUtcToLocal( source_ref_time )
            source_ref_time = datetime.strftime( source_ref_time, '%Y-%m-%d %H:%M:%S' )
        
        elif( prev_twt != '' ):#tweet
            source_ref_time = prev_twt['bloc']['local_time']


    if( use_src_ref_time is True and source_ref_time != '' ):
        #dur_glyph encodes delta between twt (current) and the tweet that was replied to or retweeted
        delta_seconds, dur_glyph = get_duration_prefix( symbols=symbols, cur_time=twt['bloc']['local_time'], prev_time=source_ref_time, blank_mark=blank_mark, minute_mark=minute_mark )
    elif( prev_twt != '' ):
        #dur_glyph encodes delta between twt (current) and prev_twt
        delta_seconds, dur_glyph = get_duration_prefix( symbols=symbols, cur_time=twt['bloc']['local_time'], prev_time=prev_twt['bloc']['local_time'], blank_mark=blank_mark, minute_mark=minute_mark )
         
    return delta_seconds, dur_glyph

def get_bloc_action_seq(symbols, twt, delta_seconds, dur_glyph, content_syntactic_seq=''):

    '''
        UPDATE get_action_glyphs() IF CHANGES
        The ACTION alphabets are
            D - Delete a tweet (moved to Change alphabet)
            f - Follow someone (moved to Change alphabet)
            L - Like a friend's tweet (moved to Change alphabet)
            l - Like a non-friend's tweet (moved to Change alphabet)
            
            P - Reply a friend (can't be checked until friendship relationship assigned)
            p - Reply a non-friend
            π - Reply self
            R - Retweet a friend (can't be checked until friendship relationship assigned)
            r - Retweet a non-friend
            ρ - Retweet self
            T - Tweet
    '''
    if( twt.get('in_reply_to_status_id', None) is not None ):

        #this is a reply
        label = symbols['non_friend_reply']['symbol']
        note = symbols['non_friend_reply']['description']
        
        if( twt.get('in_reply_to_user_id', None) == twt['user']['id'] ):
            label = symbols['self_reply']['symbol']
            note = symbols['self_reply']['description']
            
        elif( twt['bloc']['src_follows_tgt'] is True ):
            label = label = symbols['friend_reply']['symbol']
            note = note = symbols['friend_reply']['description']
            
        action_seq = {
            'seq': label, 
            'note': note, 
            'seq_dets': { 'in_reply_to_screen_name': twt.get('in_reply_to_screen_name', '') }
        }

    elif( 'retweeted_status' in twt ):

        #this is a retweet

        label = symbols['non_friend_rt']['symbol']
        note = symbols['non_friend_rt']['description']
        
        if( twt['retweeted_status']['user']['id'] == twt['user']['id'] ):
            label = symbols['self_rt']['symbol']
            note = symbols['self_rt']['description']

        elif( twt['bloc']['src_follows_tgt'] is True ):
            label = symbols['friend_rt']['symbol']
            note = symbols['friend_rt']['description']

        action_seq = {
            'seq': label, 
            'note': note, 
            'seq_dets': { 'retweet_screen_name': twt['user']['screen_name'] }
        }

    else:

        #this is a tweet 

        label = symbols['tweet']['symbol']
        note = symbols['tweet']['description']

        action_seq = {
            'seq': symbols['tweet']['symbol'], 
            'note': symbols['tweet']['description'], 
            'seq_dets': { 'user_screen_name': twt['user']['screen_name'] }
        }

    action_seq['delta_seconds'] = delta_seconds
    action_seq['seq'] = dur_glyph + action_seq['seq'] + content_syntactic_seq
    return action_seq

def get_bloc_content_sem_sent_seq(symbols, twt, txt_key, delta_seconds, dur_glyph, content_syntactic_seq=''):
    
    '''
        The SENTIMENT alphabets are
            ⋃ - Positive
            - - Neutral
            ⋂ - Negative
    '''


    exclusive_text = ''
    sent_seq = {
        'delta_seconds': -1,
        'seq': '',
        'note': '',
        'seq_dets': {}
    }

    if( txt_key in twt ):
        exclusive_text = get_twt_text_exclusively( twt[txt_key], twt['entities'] )
    
    if( exclusive_text == '' ):
        return sent_seq
    

    try:
        sent = TextBlob(exclusive_text).sentiment.polarity
    except:
        genericErrorInfo()
        return sent_seq
    
    '''
        if( sent >= 0.3333 ):    #[ 0.3333,  1.0000)
            seq = '⋃'
        elif( sent >= -0.3333 ): #[-0.3333,  0.3333)
            seq = '-'
        else:                    #[-1.0000, -0.3333)
            seq = '⋂'
    '''

    if( sent > 0 ):    
        seq = symbols['positive']['symbol']
        sent_seq['note'] = symbols['positive']['description']
    elif( sent == 0 ): 
        seq = symbols['neutral']['symbol']
        sent_seq['note'] = symbols['neutral']['description']
    else:              
        seq = symbols['negative']['symbol']
        sent_seq['note'] = symbols['negative']['description']

    
    #dur_glyph = ''
    sent_seq['delta_seconds'] = delta_seconds
    sent_seq['seq'] = dur_glyph + seq + content_syntactic_seq
    sent_seq['seq_dets'] = { 'sent': sent }
    return sent_seq

def get_bloc_change_seq(symbols, twt, prev_twt, delta_seconds, dur_glyph, fold_start_count=0, include_time=False):

    def add_delete_tweet_change(symbol, symbol_note, prev_u, cur_u, fold_start_count):

        if( 'statuses_count' not in prev_u or 'statuses_count' not in cur_u ):
            return []

        deletion_count = cur_u['statuses_count'] - prev_u['statuses_count']
        if( deletion_count > -1 ):
            return []

        deletion_count = -1 * deletion_count
        label = symbol * deletion_count
    
        return [{
            'label': label[:fold_start_count] if fold_start_count > 0 else label, 
            'note': symbol_note, 
            'details': { 'current_statuses_count': cur_u['statuses_count'], 'previous_statuses_count': prev_u['statuses_count'] }
        }]

    def add_gen_tweet_count_change(key, pos_glyph, neg_glyph, pos_label, neg_label, prev_u, cur_u, fold_start_count):

        if( key not in prev_u or key not in cur_u ):
            return []

        diff = cur_u[key] - prev_u[key]

        if( diff == 0 ):
            return []
        elif( diff > 0 ):
            label = pos_glyph * diff
            note = pos_label
        else:
            diff = (-1) * diff
            label = neg_glyph * diff
            note = neg_label

        return [{
            'label': label[:fold_start_count] if fold_start_count > 0 else label, 
            'note': note, 
            'details': { 'current_' + key: cur_u[key], 'previous_' + key: prev_u[key] }
        }]

    def add_profile_appearance_change(symbol, symbol_note, prev_u, cur_u):

        for key, val in prev_u.items():

            if( key.startswith('profile_') is False ):
                continue

            if( key not in cur_u ):
                continue

            if( prev_u[key] != cur_u[key] ):
                return [{
                    'label': symbol, 
                    'note': symbol_note, 
                    'details': { 'current_' + key: cur_u[key], 'previous_' + key: prev_u[key] }
                }]

        return []

    def add_something_change_v1(key, glyph, prev_u, cur_u):

        if( key not in prev_u or key not in cur_u ):
            return []
        
        if( prev_u[key] != cur_u[key] ):
            return [{
                'label': glyph, 
                'note': key.capitalize() + ' change', 
                'details': { 'current_' + key: cur_u[key], 'previous_' + key: prev_u[key] }
            }]

        return []

    def add_something_change_v2(key, glyph, prev_val, cur_val):
        
        if( prev_val != cur_val ):
            return [{
                'label': glyph, 
                'note': key.capitalize() + ' change', 
                'details': { 'current_' + key: cur_val, 'previous_' + key: prev_val }
            }]

        return []

    def add_geo_change(glyph, prev_u, cur_u):
        
        response = []

        #attempt to check geo-change with place
        if( 'place' in prev_u and 'place' in cur_u ):

            places = [ prev_u['place'], cur_u['place'] ]
            if( places.count(None) == 2 ):
                #both locations are None
                response = []
            
            elif( places.count(None) == 1 ):
                #at least one location is None
                response = [{
                    'label': glyph, 
                    'note': 'Place change',
                    'previous_place': None if places[0] is None else places[0],
                    'current_place': None if places[1] is None else places[1]
                }]
                
            else:
                #both locations are defined
                if( places[0]['id'] == places[1]['id'] ):
                    response = []
                else:
                    response = [{
                        'label': glyph, 
                        'note': 'Place change',
                        'previous_place': places[0],
                        'current_place': places[1]
                    }]

        if( len(response) != 0 ):
            return response
        

        #attempt to check geo-change with coordinate
        if( 'coordinate' in prev_u and 'coordinate' in cur_u ):
            
            coordinates = [ prev_u['coordinate'], cur_u['coordinate'] ]
            if( coordinates.count(None) == 2 ):
                #both locations are None
                response = []
            
            elif( coordinates.count(None) == 1 ):
                #at least one geo is None
                response = [{
                    'label': glyph, 
                    'note': 'Coordinates change',
                    'previous_coordinates': None if coordinates[0] is None else coordinates[0],
                    'current_coordinates': None if coordinates[1] is None else coordinates[1]
                }]

            else:
                #both locations are defined
                if( coordinates[0]['coordinates']['type'] == 'Point' and coordinates[1]['coordinates']['type'] == 'Point' ):                    
                    if( coordinates[0]['coordinates']['coordinates'][0] == coordinates[1]['coordinates']['coordinates'][0] and coordinates[0]['coordinates']['coordinates'][1] == coordinates[1]['coordinates']['coordinates'][1] ):
                        response = []
                    else:                        
                        response = [{
                            'label': glyph, 
                            'note': 'Coordinates change',
                            'previous_coordinates': coordinates[0],
                            'current_coordinates': coordinates[1]
                        }]
                else:
                    response = []

        return response
        
    
    '''
        The default CHANGE alphabets are
            a - Profile appearance change
            D - Delete tweet
            d - Description change
            F - Follow someone
            f - Unfollow someone
            g - Profile location change
            G - geographical location change (see coordinates and place)
            λ - Language change
            L - Liked tweet
            l - Unliked tweet
            n - Name change
            N - Handle change
            s - source change
            u - URL change
            W - Gained followers
            w - Lost followers

        fold_start_count:
            Some alphabets (e.g., D - Deletion) could lead to large labels (e.g., 1000 Ds)
            fold_start_count truncates labels to fold_start_count (this value is the same passed to bloc.util.get_bloc_variant_tf_matrix().block_variant)
        
        To do:
            test add_geo_change()
    '''
    result = {
        'delta_seconds': -1,
        'seq': '',
        'seq_dets': []
    }
    if( 'in_reply_to_status_id' not in twt or prev_twt == '' ):
        return result

    if( 'user' not in twt or 'user' not in prev_twt ):
        return result

    seq = []

    seq += add_something_change_v1( key='screen_name', glyph=symbols['handle_change']['symbol'], prev_u=prev_twt['user'], cur_u=twt['user'] )
    seq += add_something_change_v1( key='name', glyph=symbols['name_change']['symbol'], prev_u=prev_twt['user'], cur_u=twt['user'] )
    seq += add_something_change_v1( key='location', glyph=symbols['profile_location_change']['symbol'], prev_u=prev_twt['user'], cur_u=twt['user'] )
    seq += add_geo_change( glyph=symbols['geo_location_change']['symbol'], prev_u=prev_twt, cur_u=twt )
    seq += add_profile_appearance_change( symbols['profile_appearance_change']['symbol'], symbols['profile_appearance_change']['description'], prev_twt['user'], twt['user'] )
    seq += add_something_change_v1( key='url', glyph=symbols['profile_url_change']['symbol'], prev_u=prev_twt['user'], cur_u=twt['user'] )    
    seq += add_something_change_v1( key='description', glyph=symbols['description_change']['symbol'], prev_u=prev_twt['user'], cur_u=twt['user'] )
    seq += add_something_change_v2( key='source', glyph=symbols['source_change']['symbol'], prev_val=prev_twt.get('source', None), cur_val=twt.get('source', None) )
    seq += add_something_change_v2( key='lang', glyph=symbols['language_change']['symbol'], prev_val=prev_twt.get('lang', None), cur_val=twt.get('lang', None) )

    seq += add_delete_tweet_change( symbols['delete_tweet']['symbol'], symbols['delete_tweet']['description'], prev_twt['user'], twt['user'], fold_start_count )
    seq += add_gen_tweet_count_change( key='friends_count', pos_glyph=symbols['follow_someone']['symbol'], neg_glyph=symbols['unfollow_someone']['symbol'], pos_label=symbols['follow_someone']['description'], neg_label=symbols['unfollow_someone']['description'], prev_u=prev_twt['user'], cur_u=twt['user'], fold_start_count=fold_start_count )    
    seq += add_gen_tweet_count_change( key='favourites_count', pos_glyph=symbols['like_tweet']['symbol'], neg_glyph=symbols['unlike_tweet']['symbol'], pos_label=symbols['like_tweet']['description'], neg_label=symbols['unlike_tweet']['description'], prev_u=prev_twt['user'], cur_u=twt['user'], fold_start_count=fold_start_count )
    seq += add_gen_tweet_count_change( key='followers_count', pos_glyph=symbols['follower_gain']['symbol'], neg_glyph=symbols['follower_loss']['symbol'], pos_label=symbols['follower_gain']['description'], neg_label=symbols['follower_loss']['description'], prev_u=prev_twt['user'], cur_u=twt['user'], fold_start_count=fold_start_count )

    if( len(seq) == 0 ):
        return result
   
    result = {
        'delta_seconds': delta_seconds,
        'seq': ''.join([ l['label'] for l in seq ]),
        'seq_dets': seq
    }
    result['seq'] = '(' + result['seq'] + ')'
    result['seq'] = dur_glyph + result['seq'] if include_time is True else result['seq']

    return result

def get_bloc_content_syn_seq(symbols, tweet, content_syntactic_add_pause=False, txt_key='full_text', gen_rt_content=True, add_txt_glyph=True, delta_seconds=-1, dur_glyph=''):
    
    '''
        The default CONTENT-SYNTACTIC alphabets are:
            E - Media 
            H - Hashtag
            ¤ - Cashtag
            M - Mention of friend (can't be checked until friendship relationship assigned)
            m - Mention of non-friend
            q - Quote URL
            φ - Quote self
            t - Text
            U - URL
    '''

    if( gen_rt_content is False and 'retweeted_status' in tweet ):
        return {
            'seq': '',
            'seq_dets': []
        }

    if( 'retweeted_status' in tweet ):
        twt = tweet['retweeted_status']
    else:
        twt = tweet


    seq = []
    exclusive_text = ''

    if( 'extended_entities' in twt ):
        #see: http://web.archive.org/web/20210812135035/https://developer.twitter.com/en/docs/twitter-api/v1/data-dictionary/object-model/extended-entities
        if( 'media' in twt['extended_entities'] ):
            
            media = [{
                'label': symbols['media']['symbol'], 
                'note': symbols['media']['description'], 
                'details': None
            }]

            seq += media * len( twt['extended_entities']['media'] )

    if( 'entities' in twt ):
        
        for h in twt['entities'].get('hashtags', []):
            seq.append({
                'label': symbols['hashtag']['symbol'], 
                'note': symbols['hashtag']['description'], 
                'details': {'hashtag': h['text']}
            })

        for c in twt['entities'].get('symbols', []):
            seq.append({
                'label': symbols['cashtag']['symbol'], 
                'note': symbols['cashtag']['description'], 
                'details': {'cashtag': c['text']}
            })


        #assume mention is from non-friend
        if( 'user_mentions' in twt['entities'] ):

            mention_cursor = 0
            label = ''
            note = symbols['non_friend_mention']['description']

            if( twt['in_reply_to_status_id'] is None and len(twt['entities']['user_mentions']) != 0 ):
                #this is NOT a reply
                label = symbols['non_friend_mention']['symbol']

            elif( len(twt['entities']['user_mentions']) > 1 ):
                #this is a reply with a mention
                label = symbols['non_friend_mention']['symbol']
                mention_cursor = 1#skip first reply 
                

            '''
                if there are multiple mentions, recall that friendship lookup is done just once for reply or retweet
                by default retweets are (gen_rt_content=True) assigned BLOC labels
                so an checking M here will only apply to replies, and this is already done by action check, so skip M check if( twt['bloc']['src_follows_tgt'] is True ): for now
            '''

            if( label != '' ):

                for i in range(mention_cursor, len(twt['entities']['user_mentions'])):
                    
                    m = twt['entities']['user_mentions'][i]

                    #check if mention is in display text range - start
                    '''
                    #UNSURE if distinction should be made between mentions explicitly added by user vs those added by Twitter for reply chains.
                    #Twitter V1 made a distinction with display_text_range unlike V2
                    if( 'display_text_range' in twt ):
                        st, en = twt['display_text_range']
                        if( twt[txt_key][st:en].lower().find( m['screen_name'].lower() ) == -1 ):
                            #this mention is not visible in text, so skip
                            continue
                    '''
                    #check if mention is in display text range - end


                    seq.append({
                        'label': label,
                        'note': note,
                        'details': {'mention_screen_name': m['screen_name']}
                    })


        
        for u in twt['entities'].get('urls', []):

            #check for self quote - start
            u_screen_name = get_screen_name_frm_status_uri( u['expanded_url'] )

            if( u_screen_name != '' and u['expanded_url'].find('/photo/') != -1 ):
                #this is a link of an image (V2 Twitter API)
                continue
        
            if( u_screen_name == '' ):
                link_label = symbols['url']['symbol']
                link_label_note = symbols['url']['description']
            elif( u_screen_name == twt['user']['screen_name'] ):
                link_label = symbols['self_quote']['symbol']
                link_label_note = symbols['self_quote']['description']
            else:
                link_label = symbols['quote_url']['symbol']
                link_label_note = symbols['quote_url']['description']
            #check for self quote - end

            seq.append({
                'label': link_label, 
                'note': link_label_note, 
                'details': {'url': u['expanded_url']}
            })
    

        if( txt_key in twt ):
            exclusive_text = get_twt_text_exclusively( twt[txt_key], twt['entities'] )
            
            if( 'extended_entities' in twt ):
                exclusive_text = get_twt_text_exclusively( exclusive_text, twt['extended_entities'] )
            

    exclusive_text = exclusive_text.strip()
    if( exclusive_text != '' and add_txt_glyph is True ):
        seq.append({
            'label': symbols['text']['symbol'], 
            'note': symbols['text']['description'], 
            'details': {'exclusive_text': exclusive_text}
        })
    


    result = {
        'seq': ''.join([ l['label'] for l in seq ]),
        'seq_dets': seq
    }

    result['delta_seconds'] = delta_seconds
    
    if( result['seq'] != '' ):
        result['seq'] = dur_glyph + '(' + result['seq'] + ')' if content_syntactic_add_pause is True else '(' + result['seq'] + ')'

    return result

def get_bloc_content_sem_ent_seq(symbols, tweet, content_semantic_add_pause=False, gen_rt_content=True, delta_seconds=-1, dur_glyph=''):
    
    '''
        The CONTENT-SEMENTIC alphabets are:

            ENTITIES
            
            ⚇ - Person (e.g., Barack Obama, Daniel, or George W. Bush)
            ⌖ - Place (e.g., Detroit, Cali, or "San Francisco, California")
            ⋈ - Organization (e.g., Chicago White Sox, IBM)
            x - Product (e.g., Mountain Dew, Mozilla Firefox)
            ⊛ - Other (e.g., Diabetes, Super Bowl 50)
    '''

    entity_type_maps = {
        'Person': symbols['person']['symbol'],
        'Place': symbols['place']['symbol'],
        'Organization': symbols['organization']['symbol'],
        'Product': symbols['product']['symbol'],
        'Other': symbols['other']['symbol']
    }

    #by default content-se is NOT assigned to retweets
    if( gen_rt_content is False and 'retweeted_status' in tweet ):
        return {
            'seq': '',
            'seq_dets': []
        }

    if( 'retweeted_status' in tweet ):
        twt = tweet['retweeted_status']
    else:
        twt = tweet


    seq = []
    if( 'entities' in twt ):
        
        for e in twt['entities'].get('annotations', []):
            if( e['type'] in entity_type_maps ):

                seq.append({
                    'label': entity_type_maps[e['type']], 
                    'note': e['type'], 
                    'details': {'text': e['normalized_text']}
                })

 

    result = {
        'seq': ''.join([ l['label'] for l in seq ]),
        'seq_dets': seq
    }

    result['delta_seconds'] = delta_seconds
    
    if( result['seq'] != '' ):
        result['seq'] = dur_glyph + '(' + result['seq'] + ')' if content_semantic_add_pause is True else '(' + result['seq'] + ')'

    return result

def bloc_segmenter(bloc_info, created_at, local_time, segmentation_type='week_number', days_segment_count=-1):

    bloc_info['local_time'] = datetime.strftime(local_time, '%Y-%m-%d %H:%M:%S')
    bloc_info['week_number'] = format(local_time.year, '04d') + '.' + format(local_time.isocalendar()[1], '03d')


    if( segmentation_type == 'yyyy-mm-dd' ):
        bloc_info['yyyy-mm-dd'] = datetime.strftime(created_at, '%Y-%m-%d')
    
    if( days_segment_count > 0 ):
        
        bloc_info['day_of_year']     = local_time.timetuple().tm_yday
        bloc_info['day_of_year_bin'] = (bloc_info['day_of_year']-1)//days_segment_count

        bloc_info['day_of_year']     = format(local_time.year, '04d') + '.' + format(bloc_info['day_of_year'], '03d')
        bloc_info['day_of_year_bin'] = format(local_time.year, '04d') + '.' + format(bloc_info['day_of_year_bin'], '03d')
        
    bloc_info['segmentation_type'] = segmentation_type

def add_bloc_sequences(tweets, blank_mark=60, minute_mark=5, gen_rt_content=True, add_txt_glyph=True, segmentation_type='week_number', **kwargs):

    def tranfer_dets_for_stream_statuses(twt):
        
        #transfer details for tweets gotten from streams
        #See: http://web.archive.org/web/20210812135100/https://docs.tweepy.org/en/stable/extended_tweets.html
        if( 'extended_tweet' not in twt ):
            return twt

        for ky in ['full_text', 'display_text_range', 'entities', 'extended_entities']:
            if( ky in twt['extended_tweet'] ):
                twt[ky] = twt['extended_tweet'][ky]

        return twt

    def add_bloc_segments(twt, segment_id, bloc_segments):

        twt['bloc']['bloc_sequences_short'] = {}
        for dim, dim_dct in twt['bloc']['bloc_sequences'].items():
            twt['bloc']['bloc_sequences_short'][dim] = dim_dct['seq']

            #segment_id bloc segmenter - start
            bloc_segments['segments'].setdefault( segment_id, {} )

            if( segment_id > bloc_segments['last_segment'] ):
                bloc_segments['last_segment'] = segment_id

            if( dim in bloc_segments['segments'][segment_id] ):
                bloc_segments['segments'][segment_id][dim] += twt['bloc']['bloc_sequences'][dim]['seq']
            else:
                bloc_segments['segments'][segment_id][dim] = twt['bloc']['bloc_sequences'][dim]['seq']         
            #segment_id bloc segmenter - end
    
    def fmt_action_content_syntactic_bloc(symbols, bloc_segments, alph_key):

        #mv all content to behind action, e.g., go from T(qt)T(qt) to TT(qt)(qt)

        time_glyphs = [symb[1]['symbol'] for symb in symbols.items() if symb[0] != 'blank_mark']
        blank_and_time_glyphs = [symbols['blank_mark']['symbol']] + time_glyphs
        time_glyphs = ''.join(time_glyphs)

        for seg_num, segs in bloc_segments.items():
            
            if( alph_key not in segs ):
                return

            new_segs = re.split( f'([{time_glyphs}])', segs[alph_key] )
            final_segs = ''
            
            for c in new_segs:
                
                if( c in blank_and_time_glyphs ):
                    final_segs += c
                    continue

                act_glyphs = ''.join(get_action_glyphs())
                only_act = re.sub(r'\(([^\)]+)\)', '', c)
                only_cnt = re.sub(r'[()' + act_glyphs + ']', '', c)
                only_cnt = only_cnt if only_cnt == '' else f'({only_cnt})'
                final_segs += only_act + only_cnt
            
            segs[alph_key] = final_segs

    def sort_action_words(symbols, bloc_segments, alph_key):

        time_glyphs = [symb[1]['symbol'] for symb in symbols.items() if symb[0] != 'blank_mark']
        time_glyphs = ''.join(time_glyphs)
        act_glyphs = get_action_glyphs()

        for seg_num, segs in bloc_segments.items():
            if( alph_key not in segs ):
                continue
            
            new_act = ''
            prev_act = segs[alph_key]
            prev_act = re.split( f'([(){time_glyphs}])', prev_act )
            
            for w in prev_act:
                
                if( w == '' ):
                    continue

                if( w[0] in act_glyphs ):
                    new_act += ''.join( sorted(w) )
                else:
                    new_act += w
    
            segs[alph_key] = new_act

    kwargs.setdefault('keep_tweets', False)
    kwargs.setdefault('keep_bloc_segments', False)
    kwargs.setdefault('fold_start_count', 5)
    kwargs.setdefault('change_add_pause', False)
    kwargs.setdefault('time_reference', 'previous_tweet')#previous_tweet or reference_tweet
    kwargs.setdefault('sort_action_words', False)
    
    all_bloc_symbols = kwargs.get('all_bloc_symbols', {})

    if( is_symbols_good(all_bloc_symbols) is False ):
        logger.error(f'\nadd_bloc_sequences(): all_bloc_symbols is corrupt, so returning')
        return {}

    days_segment_count = kwargs.get('days_segment_count', -1)
    bloc_alphabets = kwargs.get('bloc_alphabets', ['action', 'content_syntactic', 'content_semantic_entity'])#additional valid bloc_alphabets: change, content_syntactic_with_pauses, action_content_syntactic

    segmentation_type = segmentation_type if segmentation_type in ['week_number', 'day_of_year_bin', 'yyyy-mm-dd'] else 'week_number'
    bloc_segments = {'segments': {}, 'last_segment': '', 'segment_count': 0, 'segmentation_type': segmentation_type}
    use_src_ref_time = True if kwargs['time_reference'] == 'reference_tweet' else False

    if( days_segment_count > 0 ):
        segmentation_type = 'day_of_year_bin'
        bloc_segments['segmentation_type'] = 'day_of_year_bin'
        bloc_segments['days_segment_count'] = days_segment_count

    twt_len = len(tweets)
    for i in range( twt_len ):    
        twt = tweets[i]
        twt = tranfer_dets_for_stream_statuses(twt)
        twt.setdefault('bloc', {'src_follows_tgt': None, 'tgt_follows_src': None})

        #Wed Oct 10 20:19:24 +0000 2018
        created_at = datetime.strptime(twt['created_at'], '%a %b %d %H:%M:%S %z %Y')
        bloc_segmenter( twt['bloc'], created_at, datetimeFromUtcToLocal(created_at), segmentation_type=segmentation_type, days_segment_count=days_segment_count )

    prev_twt = ''
    user_id = ''
    screen_name = ''
    tweets = sorted( tweets, key=lambda x: x['bloc']['local_time'] )
    for i in range( twt_len ):
        
        twt = tweets[i]
        segment_id = twt['bloc'][segmentation_type]
        twt_text_ky = 'full_text' if 'full_text' in twt else 'text'
        user_id = twt['user']['id']
        screen_name = twt['user']['screen_name']

        ex_txt = get_twt_text_exclusively( twt[twt_text_ky], twt['entities'] )
        
        delta_seconds, dur_glyph = get_pause(symbols=all_bloc_symbols['bloc_alphabets']['time'], twt=twt, prev_twt=prev_twt, blank_mark=blank_mark, minute_mark=minute_mark, use_src_ref_time=use_src_ref_time)
        twt['bloc']['bloc_sequences'] = {}

        if( 'action' in bloc_alphabets ):
            twt['bloc']['bloc_sequences']['action'] = get_bloc_action_seq( all_bloc_symbols['bloc_alphabets']['action'], twt, delta_seconds, dur_glyph )

        if( 'content_syntactic' in bloc_alphabets ):
            twt['bloc']['bloc_sequences']['content_syntactic'] = get_bloc_content_syn_seq(all_bloc_symbols['bloc_alphabets']['content_syntactic'], twt, content_syntactic_add_pause=False, txt_key=twt_text_ky, gen_rt_content=gen_rt_content, add_txt_glyph=add_txt_glyph, delta_seconds=delta_seconds, dur_glyph=dur_glyph)

        if( 'content_syntactic_with_pauses' in bloc_alphabets ):
            twt['bloc']['bloc_sequences']['content_syntactic_with_pauses'] = get_bloc_content_syn_seq(all_bloc_symbols['bloc_alphabets']['content_syntactic'], twt, content_syntactic_add_pause=True, txt_key=twt_text_ky, gen_rt_content=gen_rt_content, add_txt_glyph=add_txt_glyph, delta_seconds=delta_seconds, dur_glyph=dur_glyph)

        if( 'content_semantic_entity' in bloc_alphabets ):
            twt['bloc']['bloc_sequences']['content_semantic_entity'] = get_bloc_content_sem_ent_seq( all_bloc_symbols['bloc_alphabets']['content_semantic_entities'], twt, content_semantic_add_pause=False, gen_rt_content=gen_rt_content, delta_seconds=delta_seconds, dur_glyph=dur_glyph)
        
        if( 'content_semantic_sentiment' in bloc_alphabets ):
            twt['bloc']['bloc_sequences']['content_semantic_sentiment'] = get_bloc_content_sem_sent_seq( all_bloc_symbols['bloc_alphabets']['content_semantic_sentiment'], twt, txt_key=twt_text_ky, delta_seconds=delta_seconds, dur_glyph=dur_glyph, content_syntactic_seq='')

        if( 'change' in bloc_alphabets ):
            twt['bloc']['bloc_sequences']['change'] = get_bloc_change_seq( all_bloc_symbols['bloc_alphabets']['change'], twt, prev_twt, delta_seconds=delta_seconds, dur_glyph=dur_glyph, fold_start_count=kwargs['fold_start_count'], include_time=kwargs['change_add_pause'])

        if( 'action_content_syntactic' in bloc_alphabets ):
            tmp_cnt = get_bloc_content_syn_seq(all_bloc_symbols['bloc_alphabets']['content_syntactic'], twt, content_syntactic_add_pause=False, txt_key=twt_text_ky, gen_rt_content=gen_rt_content, add_txt_glyph=add_txt_glyph, delta_seconds=delta_seconds, dur_glyph=dur_glyph)
            twt['bloc']['bloc_sequences']['action_content_syntactic'] = get_bloc_action_seq( all_bloc_symbols['bloc_alphabets']['action'], twt, delta_seconds, dur_glyph, content_syntactic_seq=tmp_cnt['seq'] )

        add_bloc_segments(twt, segment_id, bloc_segments)
        prev_twt = twt
    
    if( 'action_content_syntactic' in bloc_alphabets ):
        fmt_action_content_syntactic_bloc(all_bloc_symbols['bloc_alphabets']['time'], bloc_segments['segments'], 'action_content_syntactic')
    if( 'content_syntactic_with_pauses' in bloc_alphabets ):
        fmt_action_content_syntactic_bloc(all_bloc_symbols['bloc_alphabets']['time'], bloc_segments['segments'], 'content_syntactic_with_pauses')
    
    if( kwargs['sort_action_words'] is True ):
        #must be after fmt_action_content_syntactic_bloc() is called
        sort_action_words( all_bloc_symbols['bloc_alphabets']['time'], bloc_segments['segments'], 'action' )
        sort_action_words( all_bloc_symbols['bloc_alphabets']['time'], bloc_segments['segments'], 'action_content_syntactic' )

    #measure to enable empty tweets have empty bloc - start
    #caution keys tight coupling with twt['bloc']['bloc_sequences']
    aggregate_bloc = {}
    for dim in bloc_alphabets:
        aggregate_bloc[dim] = ''
    #measure to enable empty tweets have empty bloc - start
    

    #combine bloc across multiple segments into single bloc sequence delimited by | 
    bloc_segments['segment_count'] = len(bloc_segments['segments'])
    all_segment_ids = list(bloc_segments['segments'].keys())
    all_segment_ids.sort()

    for segment_id in all_segment_ids:
        
        dimensions = bloc_segments['segments'][segment_id]
        for dim in dimensions:

            aggregate_bloc.setdefault(dim, '')
            if( dimensions[dim] == '' ):
                continue
            aggregate_bloc[dim] += dimensions[dim] + ' | '


    #remove last | which does not have bloc sequences following it
    for dim, bloc_seq in aggregate_bloc.items():
    
        aggregate_bloc[dim] = bloc_seq.strip()
        if( aggregate_bloc[dim].endswith('|') ):
            aggregate_bloc[dim] = bloc_seq.strip()[:-1]

    if( kwargs['keep_bloc_segments'] is False ):
        bloc_segments['segments'] = {}

    result = {
        'bloc': aggregate_bloc,
        'tweets': tweets if kwargs['keep_tweets'] is True else [],
        'bloc_segments': bloc_segments,
        'created_at_utc': datetime.utcnow().isoformat().split('.')[0] + 'Z',
        'screen_name': screen_name,
        'user_id': user_id,
        'bloc_symbols_version': all_bloc_symbols.get('version', '')
    }
    
    return result

def post_proc_bloc_sequences(report, seconds_mark, minute_mark, ansi_code, segmentation_type):

    if( 'tweets' not in report or 'bloc' not in report ):
        return

    tweets = report['tweets']
    aggregate_bloc = report['bloc']

    logger.info( get_timeline_request_dets(report) )
    for dim, bloc_seq in aggregate_bloc.items():

        logger.info(dim + ':')
        
        if( dim == 'action' or dim == 'content_semantic_sentiment' ):
            logger.info( color_bloc_action_str(bloc_seq, ansi_code=ansi_code) )
        else:
            logger.info(bloc_seq)
        
        logger.info('')

    logger.info( get_timeline_key_dets(seconds_mark, minute_mark, segmentation_type) )

def get_timeline_request_dets(report):

    if( 'tweets' not in report or 'bloc' not in report ):
        return ''

    if( len(report['tweets']) == 0 ):
        return ''

    tweets = report['tweets']
    details = ''
    try:
        screen_name = report['screen_name'] if report['screen_name'] != '' else tweets[0]['user']['screen_name']
        st_time = datetime.strptime(tweets[0]['bloc']['local_time'], '%Y-%m-%d %H:%M:%S')
        en_time = datetime.strptime(tweets[-1]['bloc']['local_time'], '%Y-%m-%d %H:%M:%S')
        week_diff = en_time.isocalendar()[1] - st_time.isocalendar()[1]

        delta = en_time - st_time
        delta = '(' + str(delta.days) + ' day(s), ' + str(delta.seconds//3600) + ':' + str( ((delta.seconds//60)%60) ) + '), '
        
        details = '\n@' + screen_name + '\'s BLOC for ' + str(week_diff) + ' week(s), ' + delta + str(len(tweets)) + ' tweet(s) from ' + str(st_time) + ' to ' + str(en_time)
    except:
        genericErrorInfo()
    
    return details

def get_timeline_key_dets(seconds_mark, minute_mark, segmentation_type):
    
    ky_dets = 'action key:\n'
    ky_dets += f'blank: <{seconds_mark} secs, □: <{minute_mark} mins ⚀: <hour,  ⚁: <day, ⚂: <week, ⚃: <month, ⚄: <year, ⚅: >year, | separates {segmentation_type} segments'

    return ky_dets

def get_timeline_tweets_cache_filename(cache_path, screen_name, max_pages, following_lookup, timeline_startdate, timeline_scroll_by_hours):

    cache_path = cache_path.strip()
    if( cache_path.endswith('/') is False ):
        cache_path = cache_path + '/'

    filename = ''
    try:
    
        os.makedirs( cache_path, exist_ok=True )
        
        following_lookup = '1' if following_lookup is True else '0'
        timeline_startdate = timeline_startdate.replace(' ', '_').replace(':', '-')
        timeline_scroll_by_hours = str(timeline_scroll_by_hours)

        filename = cache_path + screen_name + '_' + str(max_pages) + '_' + following_lookup + '_' + timeline_startdate + '_' + timeline_scroll_by_hours + '.json.gz'
    
    except:
        genericErrorInfo()

    return filename

def get_user_bloc(oauth_or_ostwt, screen_name, user_id='', max_pages=1, following_lookup=False, timeline_startdate='', timeline_scroll_by_hours=None, **kwargs):

    kwargs.setdefault('gen_rt_content', True)
    
    kwargs.setdefault('cache_path', '')
    kwargs.setdefault('cache_read', False)
    kwargs.setdefault('cache_write', False)

    kwargs.setdefault('blank_mark', 60)
    kwargs.setdefault('minute_mark', 5)

    kwargs.setdefault('ansi_code', '91m')
    
    tweets = []
    write_cache = True
    prev_now = datetime.now()

    if( screen_name == '' ):
        cache_filename = get_timeline_tweets_cache_filename(kwargs['cache_path'], user_id, max_pages=max_pages, following_lookup=following_lookup, timeline_startdate=timeline_startdate, timeline_scroll_by_hours=timeline_scroll_by_hours)
    else:
        cache_filename = get_timeline_tweets_cache_filename(kwargs['cache_path'], screen_name, max_pages=max_pages, following_lookup=following_lookup, timeline_startdate=timeline_startdate, timeline_scroll_by_hours=timeline_scroll_by_hours)
    

    if( kwargs['cache_path'] != '' and kwargs['cache_read'] is True ):
        
        logger.info('\nget_user_bloc() attempting to read timeline tweets from: ' + cache_filename)
        tweets = getDictFromJsonGZ( cache_filename )
        
        if( len(tweets) == 0 ):
            write_cache = True
            logger.info('\tcache MISS')
        else:
            write_cache = False
            logger.info('\tcache HIT')

    if( len(tweets) == 0 ):

        if( isinstance(oauth_or_ostwt, osometweet.OsomeTweet) ):
            tweets = v2_get_timeline_tweets(oauth_or_ostwt, user_id=user_id, max_pages=max_pages, max_results=kwargs.get('max_results', 100), timeline_startdate=timeline_startdate, timeline_scroll_by_hours=timeline_scroll_by_hours)
        else:
            tweets = get_timeline_tweets(oauth_or_ostwt, screen_name, user_id=user_id, max_pages=max_pages, following_lookup=following_lookup, timeline_startdate=timeline_startdate, timeline_scroll_by_hours=timeline_scroll_by_hours)

    if( cache_filename != '' and kwargs['cache_write'] is True and write_cache ):
        gzipTextFile(cache_filename, json.dumps(tweets, ensure_ascii=False))

    
    delta = datetime.now() - prev_now
    payload = add_bloc_sequences(tweets, **kwargs)
    
    payload['elapsed_time'] = str(delta)
    post_proc_bloc_sequences(payload, kwargs['blank_mark'], kwargs['minute_mark'], kwargs['ansi_code'], kwargs.get('segmentation_type', None))

    return payload

def is_symbols_good(all_bloc_symbols):

    if( len(all_bloc_symbols) == 0 or isinstance(all_bloc_symbols, dict) is False ):
        logger.error('\nis_symbols_good(): len(all_bloc_symbols) == 0 or isinstance(all_bloc_symbols, dict) is False, returning')
        return False

    if( 'bloc_alphabets' in all_bloc_symbols ):
        for alph in ['action', 'time', 'content_syntactic', 'change', 'content_semantic_entities', 'content_semantic_sentiment']:
            
            if( alph not in all_bloc_symbols['bloc_alphabets'] ):
                logger.error( '\nis_symbols_good(): {} alphabet not in {}, returning'.format(alph, all_bloc_symbols['bloc_alphabets'].keys()) )
                return False

            if( len(all_bloc_symbols['bloc_alphabets'][alph]) == 0 ):
                logger.error( "\nis_symbols_good(): len(all_bloc_symbols['bloc_alphabets']['{}']) == 0, returning".format(alph) )
                return False

    return True

def f1_time_function(symbols):
    
    for symb_key, symb_val in symbols.items():
        if( symb_key == 'blank_mark' ):
            continue

        symbols[symb_key]['symbol'] = '.'

def gen_bloc_for_users(screen_names_or_ids, bearer_token, consumer_key, consumer_secret, access_token, access_token_secret, max_pages=1, following_lookup=False, timeline_startdate=datetime.now().strftime('%Y-%m-%d %H:%M:%S'), timeline_scroll_by_hours=None, ngram=1, **kwargs):
    
    kwargs.setdefault('no_screen_name', False)
    bloc_symbols_file = kwargs.get('bloc_symbols_file', None)
    
    bloc_symbols_file = '{}/symbols.json'.format(os.path.dirname(os.path.abspath(__file__))) if bloc_symbols_file is None else bloc_symbols_file
    all_bloc_symbols = getDictFromFile(bloc_symbols_file)
    
    if( is_symbols_good(all_bloc_symbols) is False ):
        logger.warning(f'\ngen_bloc_for_users(): all_bloc_symbols is corrupt (bloc_symbols_file: {bloc_symbols_file}), so returning')
        return []

    if( kwargs.get('time_function', 'f1') == 'f1' ):
        f1_time_function(all_bloc_symbols['bloc_alphabets']['time'])

    kwargs['all_bloc_symbols'] = all_bloc_symbols
    if( bearer_token == '' ):
        oauth_or_ostwt = OAuth1Session(consumer_key, client_secret=consumer_secret, resource_owner_key=access_token, resource_owner_secret=access_token_secret)
    else:
        
        oauth_or_ostwt = osometweet.OsomeTweet( osometweet.OAuth2(bearer_token=bearer_token, manage_rate_limits=False) )
        if( kwargs['no_screen_name'] == False ):
            #screen_names_or_ids consists of screen_names so convert to user_ids and set kwargs['no_screen_name'] = True before calling get_user_bloc
            
            screen_names_or_ids = twitter_v2_user_lookup_ids(oauth_or_ostwt, screen_names_or_ids)
            screen_names_or_ids = [ u['id'] for u in screen_names_or_ids ]
            kwargs['no_screen_name'] = True

    i = 1
    all_users_bloc = []
    src_len = len(screen_names_or_ids)

    for scn_name_or_id in screen_names_or_ids:
        
        #logger.info(str(i) + ' of ' + str(src_len))

        user_id = ''; scn_name = ''
        if( kwargs['no_screen_name'] is True ):
            user_id = scn_name_or_id
        else:
            scn_name = scn_name_or_id

        user_bloc = get_user_bloc(
            oauth_or_ostwt=oauth_or_ostwt,
            screen_name=scn_name,
            user_id=user_id,
            max_pages=max_pages,
            following_lookup=following_lookup,
            timeline_startdate=timeline_startdate,
            timeline_scroll_by_hours=timeline_scroll_by_hours,
            **kwargs
        )
        all_users_bloc.append( user_bloc )

        i += 1

    return all_users_bloc
